//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: Prometheus_LinSysCore.C,v 1.90 2004/11/26 15:06:40 adams Exp $
//
/******************************************************************************
 * Prometheus_LinSysCore.C         Mark F. Adams                     May 2000
 *
 *               LinearSystemCore derived class for Prometheus solvers
 *   
 *     * Multigrid solvers for finite element problems in solid mechanics
 *        a) Geometric multigrid
 *        b) Smoothed aggregation algebraic multigrid 
 * 
 *     * Documentation (papers and HTML user guide) can be found on my home 
 *       page: http://www.cs.berkeley.edu/~madams
 *        
 * Limitations:
 *  1) Assumes symmetric structure of matrix. 
 * 
 * Notes,
 *  1) Parameters can be given in argument pairs "-option [value]", see the
 *     PETSc documentation for parameter specification (command line arguments)
 *  2) Constraints are handled with an "Uzawa" or Augmented Lagrange 
 *     solver that utilizes Prometheus's solvers to solve the (normalized) 
 *     stiffness matrix part of the system.  Thus, if your operator is SPD 
 *     then you can use CG even though the total linear system is indefinite 
 *     and the (unnormalized) stiffness matrix part is only symmetric positive 
 *     semi-definite.
 *  3) Geometric MG method requires elements connectivities. And currently only
 *     works with 4 node tetrahedra, 4 node quadrilateral shells and 8 node 
 *     hexahedra. See users guide for required node ordering and more details.
 * 
 ******************************************************************************/
 
#include <Prometheus_LinSysCore.h>

#include <prometheus.hh>

#if defined(PROM_USE_PETSC)
const char Prometheus_LinSysCore::version_no_[32] = "petsc.1.8.0";
#else
#error
const char Prometheus_LinSysCore::version_no_[32] = "petra.1.8.0";
#endif

static char vector_name[] = "Prometheus_Vector";
static char matrix_name[] = "Prometheus_Matrix";

// hack for a verbose flag for small objects (ie, w/o 'verbose_')
static int prom_verbose; 

#if defined(__cplusplus)
extern "C" {
#endif
  extern void exactinit();
#if defined(__cplusplus)
}
#endif

#define NOPROM_PRINT if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD, "[?]%s line %d\n", __FUNCT__, __LINE__ )
#define NOPROM_PRINT1(a) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %d\n",__FUNCT__,__LINE__,a)
#define NOPROM_PRINTD(a) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %e\n",__FUNCT__,__LINE__,a)
#define NOPROM_PRINTS(a) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %s\n",__FUNCT__,__LINE__,a)
#define NOPROM_PRINT2(a,b) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %d %d\n",__FUNCT__,__LINE__,a,b)
#define NOPROM_PRINT3(a,b,c) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %d %d %d\n",__FUNCT__,__LINE__,a,b,c)
#define NOPROM_PRINT4(a,b,c,d) if(prom_verbose>1)PetscPrintf(MPI_COMM_WORLD,"[?]%s line %d: %d %d %d %d\n",__FUNCT__,__LINE__,a,b,c,d)

#define PRINT if(verbose_>1)PetscPrintf(MPI_COMM_WORLD, "[%d]%s line %d\n",mype_, __FUNCT__, __LINE__)
#define PRINT1(a) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %d\n",mype_,__FUNCT__,__LINE__,a)
#define PRINTD(a) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %e\n",mype_,__FUNCT__,__LINE__,a)
#define PRINTS(a) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %s\n",mype_,__FUNCT__,__LINE__,a)
#define PRINT2(a,b) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %d %d\n",mype_,__FUNCT__,__LINE__,a,b)
#define PRINT3(a,b,c) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %d %d %d\n",mype_,__FUNCT__,__LINE__,a,b,c)
#define PRINT4(a,b,c,d) if(verbose_>1)PetscPrintf(MPI_COMM_WORLD,"[%d]%s line %d: %d %d %d %d\n",mype_,__FUNCT__,__LINE__,a,b,c,d)

/***********************************************************************
 *                        Prometheus_LinSysCore
 **********************************************************************/
//! class PromUzawaSolver
/*!
  class PromUzawaSolver:
*/
/* Prometheus_LinSysCore::Prometheus_LinSysCore() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::Prometheus_LinSysCore"
Prometheus_LinSysCore::Prometheus_LinSysCore( MPI_Comm comm ) :
  C_(NULL), Ctnnz_(NULL),
  lmap_(NULL),
  proj_lmap_(NULL),
  bVecArr_(NULL),
  xVec_(NULL),
  work_(NULL),
  init_guess_(NULL),
  gamma_alpha_(NULL),
  zeroIS_(NULL),
  rhsIDs_(NULL),
  proc_gnode_(NULL),
  proc_globalEq_(NULL),
  proc_lmEq_(NULL),
  proc_primEq_(NULL),
  lid_primLEq_(NULL),
  elemList_(NULL),
  geq_fixity_(NULL),
  proc_ghostCRIDs_(NULL),
  CR_id_(NULL),
  projCR_id_(NULL), 
  essbcdata_(NULL),
  othbcdata_(NULL),
  zeroEqs_(NULL),
  CRCache_(NULL),
  projCRCache_(NULL),
  prom_(NULL),
  Comm_(NULL),
  options_(NULL),
  perf_mon_(NULL),
  normRTol_(0.),
  numGlobalRows_(0),
  numGlobalRowsLm_(0),
  numGlobalRowsPrim_(0),
  currentRHS_(-1),
  dirty_guess_(FALSE),
  mg_levels_(1),
  numRHSs_(0), 
  verbose_(0),
  mype_(0),
  npe_(0), 
  num_matrices_(0),
  projNdLimit_(200),
  nSolves_(0),
  use_uzawa_(TRUE),
  matrix_setup_reuse_factor_(TRUE),
  rhs_res_scale_(FALSE),
  use_complex_(FALSE),
  MPIComm_( comm )
{
  strcpy(root_,"./");
  int tt;
  MPI_Comm_size( comm, &tt ); npe_ = tt; 
  MPI_Comm_rank( comm, &tt ); mype_ = tt; 
  
  // init exact arithmetic package (predicates.c), need once, but cheap
  exactinit();
}

/* Prometheus_LinSysCore::Prometheus_LinSysCore ***********
 *
 *  - constructor
 */
int Prometheus_LinSysCore::Init( const PromComm &comm,
				 PromOptions &opts, 
				 PromPerfMonitor &pmon )  
{
  prom_ = new Prometheus( comm, opts, pmon );
  int ierr = prom_->Init( TRUE ); CHKERRQ(ierr);
  return 0;
}

/* PromCRCache::PromCRCache() ***************************
 */
PromCRCache::PromCRCache( int geq, int len, int *geqs, double *vals ):
  len_(len), geq_(geq), geqs_(NULL), vals_(NULL), next_(NULL)
{ 
  assert(this != NULL); assert(len>0);
  int ii;
  PetscMalloc(len_*sizeof(int),&geqs_);
  for(ii=0;ii<len;ii++)geqs_[ii] = geqs[ii];
  PetscMalloc(len_*sizeof(double),&vals_);
  for(ii=0;ii<len;ii++)vals_[ii] = vals[ii];
}

PromCRCache::~PromCRCache()
{
  PetscFree(vals_);  PetscFree(geqs_);
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::parameters"
/* Prometheus_LinSysCore::parameters() ***************************
 *
 *  - Creates 'prom_' if necc.
 * 
 *  INPUT: 
 *  SIDE EFFECTS:
 */ 
int Prometheus_LinSysCore::parameters( int numParams, char** params ) 
{
  assert(this!=NULL);
  int ierr, ii, jj, done, i1, i2; 
  bool have_atol = FALSE, have_rtol = FALSE, addit, mon_init = FALSE;
  char *iname, *value, buff1[128], buff2[128], tolv[64] = "\0";
  double val;
  PRINT1(numParams); 

  // set static paramters
#if defined(PROM_HAVE_PETSC)
  if( PetscInitializeCalled == PETSC_FALSE ) {
    PETSC_COMM_WORLD =  MPIComm_;
    // can only call PetscInitialize once!
    char n1[128],*n2="-log_summary",*n3="-optionsleft"; //*n3="-out_matrix"; //*n3="-optionsleft";
    sprintf(n1,"Prometheus-%s",version_no_);
    char *n4="-trmalloc",*n5="-trmalloc_nan",*n6="-trdump",*n7="-log_info";
    char *arv[7] = {n1,n2,n3,n4,n5,n6,n7};
    char **argv = (char**)arv;
#if defined(PETSC_USE_BOPT_g)
    ii = 1; // 7 is noisy
#else 
    ii = 3;
#endif
    ierr = PetscInitialize(&ii,&argv,"./.petscrc",PETSC_NULL); CHKERRQ(ierr);
    // hack to use PETSc's communicators -- multi PETSc solvers!!!!!
    MPIComm_ = PETSC_COMM_WORLD;
#else
#error "need to have PETSc" 
    //doinit = TRUE; // petra    
#endif
  }
  else if( MPIComm_ != PETSC_COMM_WORLD ) {
#if defined(PROM_HAVE_PETSC)
    MPI_Comm_size( MPIComm_, &ii ); 
    MPI_Comm_size( PETSC_COMM_WORLD, &jj ); 
    if( ii == jj ){
      MPI_Comm_rank( MPIComm_, &ii ); 
      MPI_Comm_rank( PETSC_COMM_WORLD, &jj ); 
      if( ii == jj ) MPIComm_ = PETSC_COMM_WORLD; 
      else SETERRQ(1,"communicators changed"); 
    }
    else SETERRQ(1,"communicators changed"); // need petscized communicator!!!
#else
#endif
  }
#if defined(PROM_USE_PETSC)
  //ierr = PetscOptionsSetValue( "-mat_aij_no_inode", "");CHKERRQ(ierr);
  ierr = PetscOptionsSetValue("-sub_pc_lu_zeropivot","1.e-101");
  ierr = PetscOptionsSetValue("-sub_pc_ilu_zeropivot","1.e-101");
  ierr = PetscOptionsSetValue("-mat_aij_inode_limit","3"); CHKERRQ(ierr);
#endif
  // construct prometheus -- FEI only
  if( prom_ == NULL ) {
    assert(Comm_ == NULL);
    Comm_ = new PromComm( MPIComm_ );
    options_ = new PromOptions; 
    perf_mon_ = new PromPerfMonitor;
    prom_ = new Prometheus( *Comm_, *options_, *perf_mon_);
    mon_init = TRUE;
  }
  
  for( ii = 0 ; ii < numParams ; ii++ ) {
    for( jj = i1 = i2 = done = 0, iname = value = NULL ; 
	 params[ii][jj] != '\0' && params[ii][jj] != '\n' ; jj++ ){
      if( params[ii][jj] == '-' && iname == NULL ) {
	iname = buff1; buff1[i1++] = params[ii][jj]; // start arg1
      }
      else if( params[ii][jj] == ' ' ){
	if(iname != NULL && done==0){done++; buff1[i1++] = '\0';} // cap off arg1
	else if( value != NULL && done == 1 ){ 
	  done++; buff2[i2++] = '\0'; break; // cap off arg2
	}
	// else skip spaces
      }
      else if(iname != NULL && done == 0) buff1[i1++] = params[ii][jj]; // copy arg1
      else if( done == 1 ){ 
	if( value == NULL ) value = buff2; // start arg2
	buff2[i2++] = params[ii][jj]; // copy arg2
      }
      // incr for argc, args
      if( (params[ii][jj+1]=='\0' || params[ii][jj+1]=='\n') && 
	  done==0 && ii+1<numParams && params[ii+1][0] != '-' ) {
	if(iname != NULL){ done++; buff1[i1++] = '\0'; } // cap off arg1
	ii++; jj = -1; // spoof it
      }
    }
    if(iname != NULL && done == 0){done++; buff1[i1++] = '\0';} // cap off arg1
    if(value != NULL && done == 1){done++; buff2[i2++] = '\0';} // cap off arg2
    // call Petsc's OptionsSetValue
    if( done > 0 ) {
      if(verbose_ > 2 ){
	PetscPrintf( PETSC_COMM_WORLD,
		     "\t[%d]%s %d/%d) call OptionsSetValue(%s,%s)\n",
		     mype_,__FUNCT__,ii+1,numParams,iname,
		     (value != NULL) ? value : "" );
      }
      addit = TRUE;
      if( strcmp(iname,"-pc_type") == 0 && strcmp(value,"mg") == 0 && 
	  mg_levels_ == 1 ) { 
	mg_levels_ = 6;
      }
      else if( strcmp(iname,"-ksp_type") == 0 ) {
	if( strcmp(value,"cg") != 0 && strcmp(value,"gmres") != 0 ) {
	  // do not compute condition number if not possible
	  ierr = prom_->options_.OptionsSetValue("-prometheus_nocond",NULL); 
	  CHKERRQ(ierr);
	}
      }
      else if( strcmp(iname,"-ksp_atol") == 0 ) {
	if( value != NULL ) strcpy( tolv, value );
	have_atol = TRUE;  addit = FALSE;
      }
      else if( strcmp(iname,"-ksp_rtol") == 0 ) {
	if( value != NULL ) strcpy( tolv, value );
	have_rtol = TRUE; addit = FALSE; 
      } 
      else if( strcmp(iname,"-prom_penalty_factor") == 0 ) {
	if( value != NULL ) val = atof( value );
	else val = 10.0;
	penalty_factor_ = val; addit = FALSE; 
      }
      else if( strcmp(iname,"-prometheus_cr_tol") == 0 ) {
	addit = FALSE; 
	if( value != NULL ) {
	  val = atof( value ); cr_tol_ = val; 
	}
      }
      else if( strcmp(iname,"-ksp_tol_RHS") == 0 ){
	addit = FALSE; rhs_res_scale_ = 1; //use |b| (not |r = A * x_guess-b|) 
      }
      else if( strcmp(iname,"-ksp_tol_NONE") == 0 ) {
	if( value != NULL ) strcpy( tolv, value );
	addit = FALSE; rhs_res_scale_ = 0; 
      }        
      else if( strcmp(iname,"-out_path") == 0 ) { 
	if( value != NULL && strlen(value) < 128 ) strcpy( root_, value ); 
      }
      else if( strcmp(iname,"-max_uzawa_its") == 0 ) { 
	addit = FALSE;  
	if( value != NULL ) {
	  int vali = atoi( value );
	  if( vali > 0 ) max_uzawa_its_ = vali;
	}
      }
      else if(strcmp(iname,"-cr_pc_type") == 0){
	if( value == NULL ) addit = FALSE;
	else if( strcmp(value,"uzawa") == 0 ) {
	  use_uzawa_ = TRUE; addit = FALSE;
	}
	else {
	  use_uzawa_ = FALSE; // add it
	}
      }
      else if(strcmp(iname,"-prometheus_cr_proj_node_limit") == 0){
	addit = FALSE;
	int vali = atoi( value );
	if( vali >= 0 ) projNdLimit_ = vali;
      }
      else if( strcmp(iname,"-out_verbose") == 0 ) { 
	int vali;
	if( value != NULL ) {
	  vali = atoi( value );
	  if( vali > 0 ) verbose_ = vali;
	}
	else vali = 0;

	if( vali > 0 ) {
	  //ierr = prom_->options_.OptionsSetValue("-log_summary",NULL); 
	  if( vali > 1 ) {
	    if( vali > 2 ) {
	      ierr = prom_->options_.OptionsSetValue("-trmalloc",NULL);
	      CHKERRQ(ierr);
	      ierr = prom_->options_.OptionsSetValue("-log_info",NULL);
	      CHKERRQ(ierr);
	      ierr = prom_->options_.OptionsSetValue("-trdump",NULL); 
	      CHKERRQ(ierr);
	    }
	  }
	}
      }
      // add it in for Prometheus
      if( addit ) {
	ierr=prom_->options_.OptionsSetValue(iname,(value!=NULL) ? value : "");
	CHKERRQ(ierr);
      }
    }
  }
  // add atol
  if( tolv[0] != '\0' ) {
    // add rtol also for now
    if( have_atol && !have_rtol ) { 
      ierr = prom_->options_.OptionsSetValue("-ksp_atol",tolv);
    }
    else ierr = prom_->options_.OptionsSetValue( "-ksp_rtol", tolv ); 
    CHKERRQ(ierr);
  }

  ierr = prom_->Init( mon_init ); CHKERRQ(ierr);
  prom_verbose = verbose_;

  // hack to get 'mg' if in .petscrc
  if( mg_levels_ == 1 ) {
    bool flag; char str[32];
    ierr = prom_->options_.GetOptionString("-pc_type", str, 32, &flag);
    CHKERRQ(ierr);
    if( flag && strcmp(str,"mg") == 0 ) mg_levels_ = 8;
  }

  // 
  ierr = prom_->options_.GetOptionInt("-prometheus_levels", &ii, &addit );
  CHKERRQ(ierr);
  if( addit ) mg_levels_ = ii;
  ierr = prom_->options_.GetOptionInt("-out_verbose", &ii, &addit ); 
  CHKERRQ(ierr);
  if( addit && ii > verbose_ ) verbose_ = ii; 

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::SetRTol"
/* Prometheus_LinSysCore::SetRTol() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::SetRTol( float tol )
{
  assert(this!=NULL & prom_!=NULL && prom_->solver_!=NULL);
  int ierr = prom_->solver_->SetRTol( tol );
  return ierr;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::~Prometheus_LinSysCore"
/* Prometheus_LinSysCore::~Prometheus_LinSysCore() ***************************
 *
 *  - 
 * 
 *  INPUT: 
 *  SIDE EFFECTS:
 */
Prometheus_LinSysCore::~Prometheus_LinSysCore() 
{
  assert(this!=NULL);
  static int nobs = 0;
  PRINT1(++nobs); assert(C_ == NULL);

  if( init_guess_ != NULL ) {
    delete init_guess_; 
    assert( bVecArr_ != NULL ); assert( currentRHS_ >= 0);
    for( int ii = 0; ii < numRHSs_; ii++ ) {
      if( bVecArr_[ii] != NULL ){
	delete bVecArr_[ii]; 
      }
    }
    PetscFree( bVecArr_ );
  }
  if( xVec_ != NULL ) delete xVec_;
  if( work_ != NULL ) delete work_;
  if( uzawaC_ != NULL ) delete uzawaC_;
  else {
    PromGrid *grid0 = prom_->grid0_;
    PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(grid0->stiffness_);
    if( K != NULL && K->C_ != NULL ) RemoveCRfromAllGrids();
  }
  if( lmap_ != NULL ) delete lmap_;
  if( proj_lmap_ != NULL ) delete proj_lmap_;
  //if( penGIDs_ != NULL ) delete penGIDs_;
  if( gamma_alpha_ != NULL ) delete gamma_alpha_;
  if( proc_gnode_ != NULL ) PetscFree( proc_gnode_ );
  if( proc_globalEq_ != NULL ) PetscFree( proc_globalEq_ );
  if( proc_lmEq_ != NULL ) PetscFree( proc_lmEq_ );
  if( proc_primEq_ != NULL ) PetscFree( proc_primEq_ );
  if( rhsIDs_ != NULL ) PetscFree( rhsIDs_ ); numRHSs_ = 0; 
  if( lid_primLEq_ != NULL ) PetscFree(lid_primLEq_);
  if( elemList_ != NULL){ assert(elemList_->isEmpty()); delete elemList_; } 
  if( zeroIS_ != NULL){ PromIS is = (PromIS)zeroIS_; PromISDestroy( is ); } 
  Prometheus_ZeroEquations *zeqs = zeroEqs_ , *zeqs2;

  while( zeqs ){  
    zeqs2 = zeqs->next_; 
    for( int jj = 0; jj < zeqs->len_ ; jj++ ) { 
      assert(zeqs->idsarr_[jj] != NULL);
      PetscFree(zeqs->idsarr_[jj]); // deleting zeros
    }
    PetscFree( zeqs->zeroRowsLeq_ ); PetscFree( zeqs->sizes_ ); 
    PetscFree( zeqs->idsarr_ );
    delete zeqs; zeqs = zeqs2; 
  }
  Prometheus_EssBCData *esbc = essbcdata_, *esbc2;
  while( esbc ){ esbc2 = esbc->next_; delete esbc; esbc = esbc2; } 
  PromCRCache *crc = CRCache_, *crc2;
  while( crc ){ crc2 = crc->next_; delete crc; crc = crc2; }
  crc = projCRCache_;
  while( crc ){ crc2 = crc->next_; delete crc; crc = crc2; }
  Prometheus_OthBCData *obc = othbcdata_ , *obc2;

  while( obc ){ obc2 = obc->next_; delete obc; obc = obc2; }
  if( CR_id_ != NULL ) {
    while( !CR_id_->isEmpty() ) delete CR_id_->removeHead();
    delete CR_id_;
  }
  if( projCR_id_ != NULL ) {
    while( !projCR_id_->isEmpty() ) delete projCR_id_->removeHead();
    delete projCR_id_;
  }
  if( proc_ghostCRIDs_ != NULL ) {
    while(!proc_ghostCRIDs_->isEmpty()) delete proc_ghostCRIDs_->removeHead();
    delete proc_ghostCRIDs_;
  }
  if( geq_fixity_ != NULL ) delete geq_fixity_;

  assert(prom_ != NULL);
  delete prom_;
  if( Comm_ != NULL ) {
    delete Comm_;
    delete options_;
    delete perf_mon_;
  }
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::clone"
/* Prometheus_LinSysCore::clone() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
LinearSystemCore* Prometheus_LinSysCore::clone() 
{
  assert(this!=NULL);
  PRINT;
  return( new Prometheus_LinSysCore(MPIComm_));
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setLookup"
/* Prometheus_LinSysCore::setLookup() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::setLookup( Lookup& lookup ) 
{
  //lookup_ = &lookup;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::setUseComplex"
/* Prometheus_LinSysCore::setUseComplex() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::setUseComplex( bool b /*= true*/ )
{
  PRINT1(b);
  use_complex_ = b;
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setGlobalOffsets"
/* Prometheus_LinSysCore::setGlobalOffsets() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *   - nodeOffsets[np+1]:  [mype_] has my first node, primary eqs.
 *   - eqnOffsets[np+1]: [mype_] has my first eq.
 *   - blkEqnOffsets[np+1]: [mype_] has my first block eq (node + constriants).
 *
 *  SIDE EFFECTS:
 *   - sets 'proc_globalEq_' 'proc_gnode_' 'proc_lmEq_' 'proc_primEq_'
 */
int Prometheus_LinSysCore::setGlobalOffsets( int len, 
					     const int* nodeOffsets, // proc_gnode
					     const int* eqnOffsets,
					     const int* blkEqnOffsets ) 
{
  int ii,ierr;
  PRINT3(nodeOffsets[len-1],eqnOffsets[len-1],blkEqnOffsets[len-1]);
  assert(len == npe_+1);

  if( verbose_ > 3 || (mype_==0 && verbose_ == 3) ){
    PromContext::printArr(stderr,nodeOffsets,len,"nodeOffsets[%d]");
    PromContext::printArr(stderr,eqnOffsets,len,"eqnOffsets[%d]");
    PromContext::printArr(stderr,blkEqnOffsets,len,"blkEqnOffsets[%d]");
  }

  // 'proc_globalEq_'
  if( proc_globalEq_ == NULL ) {
    ierr = PetscMalloc(sizeof(int)*len,&proc_globalEq_);  CHKERRQ(ierr);
  }
  for( ii=0 ; ii<len ; ii++ ) proc_globalEq_[ii] = eqnOffsets[ii];
  numGlobalRows_ = eqnOffsets[npe_];
  
  // 'proc_gnode_'
  if( proc_gnode_ == NULL ) {
    ierr = PetscMalloc(sizeof(int)*len,&proc_gnode_);  CHKERRQ(ierr);
  }
  for( ii=0 ; ii<len ; ii++ ) proc_gnode_[ii] = nodeOffsets[ii];
  if(npe_+1!=len){PRINT2(npe_,len-1);SETERRQ(1,"np+1!=len");}

  // 'proc_lmEq_'
  if( proc_lmEq_ == NULL ) {
    ierr = PetscMalloc( sizeof(int)*len,&proc_lmEq_);  CHKERRQ(ierr);
  }
  for(ii=0 ; ii<len; ii++) proc_lmEq_[ii] = blkEqnOffsets[ii]-proc_gnode_[ii];
  numGlobalRowsLm_ = proc_lmEq_[npe_]; assert(CR_id_== NULL);
  if( numGlobalRowsLm_ > 0 ) { 
    int t = 4*(proc_lmEq_[mype_+1]-proc_lmEq_[mype_])/3 + 3;
    CR_id_= new PromTable<PromCR*>(t);
  }

  //  
  if( proc_primEq_ == NULL ) {
    ierr = PetscMalloc(sizeof(int)*len,&proc_primEq_);  CHKERRQ(ierr);
  }
  for(ii=0 ; ii<len ; ii++) proc_primEq_[ii] = eqnOffsets[ii]-proc_lmEq_[ii];
  numGlobalRowsPrim_ = proc_primEq_[npe_];

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::putNodalFieldData"
/* Prometheus_LinSysCore::putNodalFieldData() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::putNodalFieldData( int fieldID, 
					      int fieldSize,
					      int* nodeNumbers, 
					      int numNodes,
					      const double* data )
{ 
  PRINT3(fieldID,fieldSize,numNodes);
  PromGrid *grid0 = prom_->grid0_;
  if( grid0 == NULL ) SETERRQ( 1, "grid0 == NULL" ); 

  if( fieldID == -5555 ) { // not used
    // reorder of G-S for a direction with 'wight' (in 'data')
    if( fieldSize != 1 ) SETERRQ( 1, "fieldSize != 1" ); 
    PromSolver *solver = prom_->solver_; 
    if( solver == NULL ) SETERRQ( 1, "solver == NULL" ); 
    // get solvers
    PromPCGaussSeidel *gspc = NULL;
    PromPCMG *mgpc = dynamic_cast<PromPCMG*>(solver->pc_);
    if( mgpc != NULL ) {
      PromMG *mg = mgpc->mgstack_;
      gspc = dynamic_cast<PromPCGaussSeidel*>(mg->presmooth_->pc_); 
    }
    else gspc = dynamic_cast<PromPCGaussSeidel*>(solver->pc_);
    // doit
    if( gspc != NULL ) {
      // reorder fine grid
      double *pv, wht; int ierr,xx,nn; assert(bVecArr_!=NULL&&currentRHS_>=0);
      PromVector *vec = bVecArr_[currentRHS_]; assert(vec!=NULL); 
      const int my0 = proc_gnode_[mype_], nloc = grid0->nLocalNd_;
      assert(nloc == vec->getn()); 
      // insert weights
      for( xx = nn = 0 ; xx < numNodes ; xx++ ){
	int gid = nodeNumbers[xx], lid = gid - my0; wht = data[xx];
	if( lid >= 0 && lid < nloc ) {
	  ierr = vec->SetValues(1, &gid, &wht, INSERT_VALUES); CHKERRQ(ierr);
	  nn++;
	}
      }
      if(nn != nloc)SETERRQ( 1, "too few locals provided" ); 

      // insert weights
      ierr = vec->GetArray( &pv );    CHKERRQ(ierr);
      ierr = gspc->ReorderNodes( pv, nloc ); CHKERRQ(ierr);
      ierr = vec->RestoreArray( &pv );    CHKERRQ(ierr);
      // coarse grids
      if( mgpc != NULL ) {
	PromMG *mg = mgpc->mgstack_, *cmg = mg->next_;
	PromGrid *grid = grid0;
	while( mg->next_ != NULL ){
	  cmg = mg->next_; PromGrid *nextg = grid->nextgrid_;   
	  gspc = dynamic_cast<PromPCGaussSeidel*>(cmg->presmooth_->pc_); 
	  if( gspc == NULL ) break; // no more g-s (coarse grid)
	  PromVector *nextvec = dynamic_cast<PromVector*>(cmg->r_);
	  assert(nextvec!=NULL);
	  PromMatrix *PP = grid->prolongator_; assert(PP != NULL);
	  // project weights
	  ierr = PP->PromMatrix::MultTranspose(vec, nextvec);  // v2 <= R * v1
	  CHKERRQ(ierr);
	  // insert weights
	  ierr = nextvec->GetArray( &pv );    CHKERRQ(ierr);
	  ierr = gspc->ReorderNodes( pv, nextvec->getn() ); CHKERRQ(ierr);
	  ierr = nextvec->RestoreArray( &pv );    CHKERRQ(ierr);
	  // increment
	  grid = nextg; vec = nextvec; mg = cmg;
	}
      }
    }
    // else nothing to do...
  }
  else { // coords
    //FILE *file;file=fopen("coords","w"); 
    //PromContext::printArr(file,data,3*numNodes,"coords[%d]",3);
    //fclose(file);
    putCoords_private( fieldID, fieldSize, nodeNumbers, numNodes, data,
		       prom_->right_ );
  }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setConnectivities"
/* Prometheus_LinSysCore::setConnectivities() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */ 
int Prometheus_LinSysCore::setConnectivities( GlobalID elemBlock,
					      int numElements,
					      int numNodesPerElem,
					      const GlobalID* elemIDs,
					      const int* const* connNodes )
{
  if(verbose_>3)PRINT3(elemBlock,numElements,numNodesPerElem);

  const int nen1 = numNodesPerElem + 1;
  if(elemList_ == NULL) elemList_ = new PromList<int*>(128);
  int ii, ierr, *arr;
  for( ii = 0  ; ii < numElements ; ii++ ) {
    const int *ids = connNodes[ii];
    ierr = PetscMalloc(nen1*sizeof(int),&arr); CHKERRQ(ierr);
    elemList_->addTail( arr ); // [ n1, n2, ... , nn, -matid ]
    for(int jj=0 ; jj < numNodesPerElem ; jj++ ) *arr++ = ids[jj];
    *arr++ = -(elemBlock+1); // negative as sentinal
    if(verbose_>4 || (mype_==0 && verbose_ > 3)){
      PromContext::printArr(stderr,arr-nen1,nen1,"elem[%d]");
    }
  }

  return 0; 
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::eq_map_private"
/* Prometheus_LinSysCore::eq_map_private() ***************************
 *
 *  INPUT: 
 *    - geq: gei equation number
 *    - myeq: (out)
 *    - lamb: (out)
 *    - pproc: (out)
 *
 *  SIDE EFFECTS:
 */
inline int Prometheus_LinSysCore::eq_map_private( const int geq, 
						  int *myeq, 
						  bool *lamb,  
						  int *pproc ) const 
{
  int proc; 
  *pproc = -1;
#if defined(PETSC_USE_BOPT_g)
  if( geq < 0 || geq >= numGlobalRows_ ) {
    printf("\tERROR: geq(%d) < 0 || geq >= numGlobalRows_(%d)\n",geq,numGlobalRows_);
    return -1;
  }
#endif 
  // get proc
  float fproc = (float)geq * (float)npe_ / (float)numGlobalRows_;
  proc = (int)(fproc + .5); 
  if( proc > npe_ ) proc = npe_;
  while( geq < proc_globalEq_[proc] || geq >= proc_globalEq_[proc+1] ) {
    if(  geq < proc_globalEq_[proc] ) proc--;
    else proc++;
  }
#if defined(PETSC_USE_BOPT_g)
  if( proc<0 || proc >= npe_ ) {
    printf("\tERROR:  proc(%d)<0 || proc >= npe_(%d)\n",proc,npe_);
    return -1;
  }
#endif
  *pproc = proc;
  if( proc_lmEq_ == NULL ){ *lamb = false; *myeq = geq; } // no LMs
  else {
    int nPrim,prim0,eq0,localEq; 
    eq0 = proc_globalEq_[proc];
    prim0 = proc_primEq_[proc];
    nPrim = proc_primEq_[proc+1] - prim0; 
    // map
    localEq = geq - eq0;
    if( localEq < nPrim ){ // Primary eq 
      *lamb = false; 
      *myeq = prim0 + localEq; 
    }
    else{ // LM
      *lamb = true;
      int lm0 = proc_lmEq_[proc];
      *myeq = lm0 + (localEq - nPrim); // stsrtLm + localLm
    }
  }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::id_map_private"
/* Prometheus_LinSysCore::id_map_private() ***************************
 *
 *  INPUT: 
 *    - gid: gei iduation number
 *    - mygid: (out)
 *    - lamb: (out)
 *    - pproc: (out)
 *
 *  SIDE EFFECTS:
 */
inline int Prometheus_LinSysCore::id_map_private( const int gid, 
						  int *mygid, 
						  bool *lamb,  
						  int *pproc ) const 
{
#if defined(PETSC_USE_BOPT_g)
  if(verbose_>3) PRINT1(gid); // this is too noisy
#endif
#define PROC_BLKEQOFFSET(i) ( (i<0) ? -999999 : (proc_lmEq_[i]+proc_gnode_[i]) )
  int proc,id0,numGlobalNodes=PROC_BLKEQOFFSET(npe_); 
  *pproc = -1;
#if defined(PETSC_USE_BOPT_g)
  if( gid < 0 || gid >= numGlobalNodes ){
    PRINT3(gid,npe_,numGlobalNodes);
    SETERRQ(1,"gid >=0 && gid < proc_gid[numprocs_]"); 
  }
#endif 
  // get proc
  float fproc = (float)gid * (float)npe_ / (float)numGlobalNodes;
  proc = (int)(fproc + .5); 
  // blkEqnOffsets = proc_lmEq_ + proc_gnode_
  if( proc > npe_ ) proc = npe_;
  while(gid < (id0=PROC_BLKEQOFFSET(proc)) || gid >= PROC_BLKEQOFFSET(proc+1)){
    if( gid < id0 ) proc--;
    else proc++;
  }
#if defined(PETSC_USE_BOPT_g)
  if(proc<0 || proc >= npe_){
    SETERRQ(1,"proc < 0 || proc >= npe_");
  }
#endif
  *pproc = proc;
  if( proc_lmEq_ == NULL ){ *lamb = FALSE; *mygid = gid; } // no LMs
  else {
    int nPrim,prim0,id0,lid; 
    id0 = PROC_BLKEQOFFSET(proc);
    prim0 = proc_gnode_[proc];
    nPrim = proc_gnode_[proc+1] - prim0; 
    // map
    lid = gid - id0;
    if( lid < nPrim ){ // Primary id 
      *lamb = FALSE; 
      *mygid = prim0 + lid; 
    }
    else{ // LM
      *lamb = TRUE;
      int lm0 = proc_lmEq_[proc];
      *mygid = lm0 + (lid - nPrim); // stsrtLm + localLm
    }
  }
  
  return 0;
}
 
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::resetConstraints"
/* Prometheus_LinSysCore::resetConstraints() ***************************
 *
 *  - zero constraint matrix.
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::resetConstraints( double vv ) 
{
  PRINT1(numGlobalRowsLm_);
  
  bool flg;
  prom_->options_.HasOptionName( "-prometheus_resetCR", &flg );  
  if( flg ){ 
    const int *ids = PromContext::prom_logs_,lres=ids[LOG_FEI_RESETCR];
    const int lpart=ids[LOG_PARTITION];
    if(vv != 0.0) SETERRQ1(1, "scale = %g", vv ); 
    //if( have_pen_ ) return 1; 
    if( numGlobalRowsLm_ == 0 ){
      assert(uzawaC_==NULL);
      assert(geq_fixity_==NULL||geq_fixity_->isEmpty());
      return 1; // penalty not supported
    }
    int ierr;
    PromGrid *grid0 = prom_->grid0_;
    PromMatrix *KK = grid0->stiffness_;
        
    // " FEI: resetCR"  
    prom_->perf_mon_.EventBegin(lres);
    // " FEI: BCs & reg."
    prom_->perf_mon_.EventBegin(lpart);
    
    if( uzawaC_ != NULL ) { // Sierra can get in a strange state!
      assert( use_uzawa_ ); 
      if( !uzawaC_->isAssembled() ){
	prom_->perf_mon_.EventEnd(lpart);
	prom_->perf_mon_.EventEnd(lres);
	SETERRQ1(-1,"***C not assembled, matrixLoadComplete !called, nCR=%d",
		 CR_id_==NULL ? -1 : CR_id_->getCount() );
      }
#if defined(PETSC_USE_BOPT_g)
      double beta, beta2;
      Pen_stiff_->NormMax( &beta );
#endif
      // subtract out reg. and delete C
      ierr = setRegularization_private(KK,uzawaC_,diag_,-penalty_factor_);
      CHKERRQ(ierr);
#if defined(PETSC_USE_BOPT_g)
      Pen_stiff_->NormMax(&beta2); 
      if(beta!=beta2)PetscPrintf(MPIComm_,
				 "\t** Prometheus Warning: line %d %g != %g\n",
				 __LINE__,beta,beta2);
#endif
      ierr = KK->Assembly();        CHKERRQ(ierr);
      KK->SetDirty( FALSE ); // this is required, but !true

      // delete old contraints
      delete uzawaC_; uzawaC_ = NULL; assert(grid0->CRmap_ == NULL);
      assert( lmap_ != NULL );
      delete lmap_; lmap_ = NULL; // we take care of 'lmap_' in Uzawa!
      if( proj_lmap_ != NULL ) {
	delete proj_lmap_;
	proj_lmap_ = NULL; 
	assert(prom_->solver_ != NULL);
	ierr = prom_->solver_->projector_->ClearProjC();CHKERRQ(ierr);
      }
      else assert(prom_->solver_->projector_ == NULL);
    }
    else if( !use_uzawa_ ) {
      assert(proj_lmap_ == NULL);
      // KKT_MG remove all traces of old contraints
      int ierr = RemoveCRfromAllGrids();  CHKERRQ(ierr);  
      assert( proj_lmap_ == NULL );
    }
    assert( !KK->isDirty() );
    assert( CR_id_ == NULL || CR_id_->isEmpty() );
    assert( projCR_id_ == NULL || projCR_id_->isEmpty() );
    
    // this needs to be reset every new CR
    if( geq_fixity_ != NULL ) geq_fixity_->removeAlld();

    prom_->perf_mon_.EventEnd(lpart);
    prom_->perf_mon_.EventEnd(lres);
    return 0;
  }
  else {
    PRINTS(" -- skip --");
    //assert( !prom_->grid0_->stiffness_->isDirty() );
    assert( CR_id_ == NULL || CR_id_->isEmpty() );
    assert( projCR_id_ == NULL || projCR_id_->isEmpty() );
    // this needs to be reset every new CR
    if( geq_fixity_ != NULL ) geq_fixity_->removeAlld();
    return 1; // disabled
  }
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::setMultCREqns"
/* Prometheus_LinSysCore::setMultCREqns() ***************************
 *
 *  - get nodes of CR equations, collect needed ghost nodes
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::setMultCREqns( int multCRSetID,
					  int numCRs, 
					  int numNodesPerCR,
					  int** nodeNumbers, /*in my ID space!*/
					  int** /*eqnNumbers*/,
					  int* /*fieldIDs*/,
					  int* multiplierEqnNumbers )
{
  if(verbose_>3||(verbose_==2&&multCRSetID==0))PRINT2(numCRs,numNodesPerCR); 
  int ii,jj,*nds,tablesz = (4*(proc_lmEq_[mype_+1]-proc_lmEq_[mype_])/3+17);

  // keep track for extraedges
  for( ii = 0 ; ii < numCRs ; ii++ ) { 
    PromCR *mcr = new PromCR( numNodesPerCR );
    CHKNEW(mcr,1);
    int ierr,myCReq,proc,gid,CRgeq=multiplierEqnNumbers[ii]; bool lamb;
    ierr = eq_map_private( CRgeq, &myCReq, &lamb, &proc ); CHKERRQ(ierr);
    if( proc != mype_ ){ SETERRQ(1,"proc != myproc_"); }
    if( !lamb ){ SETERRQ(1,"!Lagrange mult eq");}
    mcr->setEq( myCReq );
    for( jj = 0, nds = nodeNumbers[ii] ; jj < numNodesPerCR ; jj++ ){
      gid = nds[jj]; mcr->gnodes_[jj] = gid;
      PromGrid::FindProc( proc_gnode_, npe_ , gid, proc );      
      // cache needed CR ghosts eqs here
      if( proc != mype_ ) {
	if(geq_fixity_==NULL) geq_fixity_ = new PromTable<double>(tablesz);
	geq_fixity_->addd( -999.0, gid + 1 ); // cache (count) neede eqs here
      }
    }
    // seperate langrange and penalty constraints 
    if( numNodesPerCR < projNdLimit_ || !use_uzawa_) {
      CR_id_->add( mcr, myCReq + 1 );
    }
    else { 
      // large CR in Uzawa - project
      if( projCR_id_ == NULL ) projCR_id_ = new PromTable<PromCR*>(6);
      projCR_id_->add( mcr, myCReq + 1 );  
    }
  }

  return 0; 
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setPenCREqns"
/* Prometheus_LinSysCore::setPenCREqns() ***************************
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::setPenCREqns(int /*penCRSetID*/, int numCRs, 
					int numNodesPerCR, int** nodeNumbers, 
					int** /*eqnNumbers*/,int* /*fieldIDs*/)
{
  if( verbose_ > 3 ) PRINT2(numCRs,numNodesPerCR);
  /*if( numCRs > 0 ) have_pen_ = TRUE;
    int ierr,ii,jj,gid,proc; bool lam;
    if( penGIDs_ == NULL ) {
    int tsz= (numCRs>17) ? 4*numCRs/3 : 117;
    penGIDs_ = new PromTable<int>(tsz);
    }
    for( ii = 0 ; ii < numCRs ; ii++ ){
    int *ipr = nodeNumbers[ii];
    for(jj=0;jj<numNodesPerCR;jj++){
    int feigid = ipr[jj]; 
    ierr = id_map_private( feigid, &gid, &lam, &proc ); CHKERRQ(ierr);
    if(lam)SETERRQ1(1, "lamb: feigid=%d",feigid);
    penGIDs_->add(gid + 1);
    }
    }
    }*/
    
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setMultCRComplete"
/* Prometheus_LinSysCore::setMultCRComplete() **************************
 *
 *  - create 'C' (and its map) and 'Pen_stiff_'. 
 *  - remake p vectors on 2nd pass (not used)
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *    - creates 'C' and 'Pen_stiff_'
 *    - if Uzawa: 'C_' gets ProjC 
 *    - if "all-at-once", on first pass, 'C_' gets C 
 */
int Prometheus_LinSysCore::setMultCRComplete() 
{
  if( numGlobalRowsLm_ == 0 ) return 0;
  const int my0lm = proc_lmEq_[mype_], nlocLm = proc_lmEq_[mype_+1] - my0lm;
  const int my0eq=proc_primEq_[mype_],nlocPrims=proc_primEq_[mype_+1]-my0eq;
  const int my0=proc_gnode_[mype_],next0=proc_gnode_[mype_+1];
  int ierr,gid,ncr,ii,ndfAv=nlocPrims/(next0-my0);
  MPI_Comm wcomm = MPIComm_;                         assert(CR_id_ != NULL);
  PromMatrix *CC = NULL; TablePos tpos;
  int *Cnz=NULL,*CnzOff=NULL,*Ctnz=NULL,logid=LOG_FEI_INIT_STR;
  PRINT2(proc_lmEq_[npe_], CR_id_->getCount());

  // reality check
  PetscFunctionBegin;
  if( proc_globalEq_ == NULL ){SETERRQ(1,"proc_globalEq_ == NULL");}  
  if( proc_globalEq_[npe_] != numGlobalRows_ ){SETERRQ(1,"ng!=pg[np]");}  
  if( PROM_BSZ < npe_ ){ SETERRQ(1,"PROM_BSZ < np");}  

  // " FEI: init. str."
  prom_->perf_mon_.EventBegin(PromContext::prom_logs_[logid]);

  // construct C matrix, and get regularization edges  
  if( nlocLm > 0 ){
    // on nnz
    ierr = PetscMalloc(nlocLm*sizeof(int),&Cnz); CHKERRQ(ierr);
    ierr = PetscMalloc(nlocLm*sizeof(int),&CnzOff); CHKERRQ(ierr);
    for( ii = 0 ; ii < nlocLm ; ii++ ) CnzOff[ii] = Cnz[ii] = 0;
  }

  if( !use_uzawa_ && nlocPrims > 0 ) {
    ierr = PetscMalloc( nlocPrims * sizeof(int), &Ctnz ); CHKERRQ(ierr);
    for( ii = 0 ; ii < nlocPrims ; ii++ ) Ctnz[ii] = 0;
  }
  // 
  tpos = CR_id_->getHeadPosition();
  for( ncr = 0 ; tpos ; ncr++ ) {
    PromCR* mcr = CR_id_->getNext( tpos );
    int *pi=mcr->gnodes_,len=mcr->len(),geq=mcr->eq(),leqi=geq-my0lm; 
    assert(leqi >= 0 && leqi < nlocLm);
    // count locals
    for( ii = 0 ; ii < len ; ii++ ) {
      gid = pi[ii];
      if( (gid >= my0 && gid < next0) || !use_uzawa_ ) {
	Cnz[leqi] += ndfAv;
	if( Ctnz != NULL ) {
	  for(int jj=0, eq=(gid-my0)*ndfAv ; jj<ndfAv ; jj++, eq++) {
	    if(eq >= 0 && eq < nlocPrims) Ctnz[eq]++;
	  }
	}
      }
      else CnzOff[leqi] += ndfAv;
    }
  }
  if( Ctnz != NULL ) {
    for(ii=0 ; ii<nlocPrims ; ii++) if(Ctnz[ii]>nlocLm) Ctnz[ii] = nlocLm;
  }
  // need to count projection CRs also, keep dummy equations in 'usawaC_'
  if( projCR_id_ != NULL ) ncr += projCR_id_->getCount();
  // send/recv lists, add to Cnz and Coffnz, and set prom. reg. data str.
  PromGrid *grid0 = prom_->grid0_;
  const bool is1stPass = (grid0 == NULL) ? TRUE : FALSE;
  PromMap *lmap =  new PromMap( ncr, 1, prom_->Comm_ );

  lmap_ = lmap;
  CC = new PromMatrix(*lmap,0,Cnz,0,CnzOff,nlocPrims,numGlobalRowsPrim_);
  if(CC==NULL || !CC->isOK()) SETERRQ(1,"new PromMatrix failed");

  // get projection
  if( !use_uzawa_ ) { // KKT-MG solves    
    assert(projCR_id_ == NULL); assert(CC != NULL);
    if( is1stPass ) C_ = CC;  // cache C for latter 'KK->C'
    else{
      PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(grid0->stiffness_);
      assert(K!=NULL); K->C_ = CC;
      grid0->CRmap_ = lmap_; // done in 'setMatrixStructure' in first pass
    }
    // create C'
    if( is1stPass ) Ctnnz_ = Ctnz;
    else{
      assert(grid0->scalarMap_ != NULL);  
      CC = new PromMatrix( *grid0->scalarMap_, 0, Ctnz, 6, NULL,
			   nlocLm, numGlobalRowsLm_ );
      PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(grid0->stiffness_);
      PetscFree(Ctnz); assert( K != NULL && K->Ct_ == NULL );
      K->Ct_ = CC; 
    }
    CC = NULL;
  }
  else {
    // set uzawa
    assert(uzawaC_ == NULL);  uzawaC_ = CC;
    if( !is1stPass && uzawaC_ != NULL ){ 
      // PromUzawaSolver code! -- redo 'pwork_'
      ierr = PromUzawaSolver::Reset( *lmap ); CHKERRQ(ierr);
    }
    // projection CRs
    int jj = (projCR_id_ == NULL) ? 0 : projCR_id_->getCount() ;
    MPI_Allreduce( &jj, &ii, 1, MPI_INT, MPI_SUM, wcomm );
    if( ii > 0 ) {
      // have projection operator -- construct it
      for( ii = 0 ; ii < nlocLm ; ii++ ) CnzOff[ii] = Cnz[ii] = 0;
      if ( projCR_id_ != NULL ) {
	tpos = projCR_id_->getHeadPosition();
	for( ncr = 0 ; tpos ; ncr++ ) {
	  PromCR* mcr = projCR_id_->getNext( tpos ); 
	  int *pi = mcr->gnodes_, len = mcr->len(), leqi = mcr->eq()-my0lm;
	  assert(leqi >= 0 && leqi < nlocLm);
	  // count locals
	  for( jj = 0 ; jj < len ; jj++ ) {
	    gid = pi[jj];
	    if(gid < my0 || gid >= next0) CnzOff[leqi] += ndfAv;
	    else Cnz[leqi] += ndfAv;
	  }
	}
      }
      PromMap *plmap = new PromMap( ncr, 1, prom_->Comm_ );
      PromMatrix *prCC;
      prCC= new PromMatrix(*plmap,0,Cnz,0,CnzOff,nlocPrims,numGlobalRowsPrim_);
      proj_lmap_ = plmap;  
      C_ = prCC;             // cache projC
    }
  } 
  // 
  if( Cnz != NULL ) { PetscFree(Cnz); PetscFree(CnzOff); }

  PromCRVector *XX = dynamic_cast<PromCRVector*>(xVec_);
  if( XX != NULL && XX->p_ != NULL ) { // second pass vecs
    assert(0); // not used (?) used for 'resetCRs'
    PromVector *v; assert(lmap != NULL);
    PromCRVector *WW = dynamic_cast<PromCRVector*>(work_);
    assert(WW != NULL && WW->p_ != NULL);
    PromCRVector *IG = dynamic_cast<PromCRVector*>(init_guess_);
    assert(WW != NULL && WW->p_ != NULL);
    delete XX->p_; delete WW->p_; delete IG->p_;
    v = new PromVector( *lmap ); XX->p_ = v; v->Set( 0.0 );
    v = new PromVector( *lmap ); WW->p_ = v; v->Set( 0.0 );
    v = new PromVector( *lmap ); IG->p_ = v; v->Set( 0.0 );
    for( ii = 0; ii < numRHSs_; ii++ ) {
      PromCRVector *BB = dynamic_cast<PromCRVector*>(bVecArr_[ii]);
      assert(BB != NULL && BB->p_!=NULL); delete BB->p_;
      v = new PromVector( *lmap ); BB->p_ = v; v->Set( 0.0 );
    }
  }
  prom_->perf_mon_.EventEnd(PromContext::prom_logs_[logid]);

  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setMatrixStructure"
/* Prometheus_LinSysCore::setMatrixStructure() ***************************
 *
 *  - FEI initializer
 *
 *  INPUT: 
 *   - ptColIndices
 *   - ptRowLengths
 *   - blkColIndices
 *   - blkRowLengths
 *   - ptRowsPerBlkRow
 *   - mat: PETSc Mat object - fine grid matrix (used in PETSc PC interface)
 *
 *  SIDE EFFECTS:
 *   - creates, and adds into, 'proc_ghostCRIDs_'
 */ 
int Prometheus_LinSysCore::setMatrixStructure( int** /*ptColIndices*/, 
					       int* /*ptRowLengths*/,
					       int** blkColIndices, 
					       int* blkRowLengths,
					       int* ptRowsPerBlkRow,
					       void *mat /*= NULL*/ )
{
  //
  //This function is where we establish the structures/objects associated
  //with the linear algebra library. i.e., do initial allocations, etc.
  int       tt,ii,jj,kk,ghostlid,ierr,lid,proc,gid,xx;
  const int my0=proc_gnode_[mype_], next0=proc_gnode_[mype_+1];
  const int nLocNd = next0 - my0; PromTable<int> *table;
  const int tsz = nLocNd/2 + 17, tag=prom_->Comm_.getNewTag();
  PromTable<PromTable<int>*> proc_bounds(31);
  PromTable<int> gid_ghostLid(tsz);
  const int reqsz = sizeof(MPI_Request)/sizeof(int);
  PromList<MPI_Request*> smessages, rmessages; 
  MPI_Comm wcomm = MPIComm_;
  MPI_Status status;
  PRINT2(numGlobalRowsPrim_,numGlobalRowsLm_);

  // reality check
  PetscFunctionBegin;
  if( proc_globalEq_ == NULL ){SETERRQ(1,"proc_globalEq_ == NULL");}  
  if( proc_globalEq_[npe_] != numGlobalRows_ ){SETERRQ(1,"ng!=pg[np]");}  
  if( PROM_BSZ < npe_ ){SETERRQ(1,"PROM_BSZ < np");}

  prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FEI_INIT_STR]);
  // determine if constant ndf
  int s2[2],r2[2];
  if( nLocNd > 0 ) {
    for(ii=1,tt=ptRowsPerBlkRow[0];ii<nLocNd;ii++) {
      if(ptRowsPerBlkRow[ii] != tt) break;
    }
    if( ii != nLocNd ) s2[0] = tt = 0; // broke
    else s2[0] = tt; // high / low
  }
  else{ s2[0] = 0; tt = 999; }
  s2[1] = (elemList_!=NULL) ? 1 : 0;
  MPI_Allreduce(&tt, &jj, 1, MPI_INT, MPI_MIN, wcomm );
  MPI_Allreduce(s2, r2, 2, MPI_INT, MPI_MAX, wcomm );
  const bool constndf = (jj==r2[0] && jj>0); 
  const int ndf = constndf ? jj : 0;
  const bool haveelems = (r2[1]!=0);
  // construct 'proc_ghostCRIDs_'
  for( xx = 0 ; xx < nLocNd ; xx++ ) {
    int *indices = blkColIndices[xx], len = blkRowLengths[xx];
    for( jj = 0 ; jj < len ; jj++ ) {
      int feigid = indices[jj], proc; bool lamb;
      ierr = id_map_private( feigid, &gid, &lamb, &proc ); CHKERRQ(ierr);
      if( proc != mype_ ) { 
	if( lamb ) { 
	  if( proc_ghostCRIDs_ == NULL ) {
	    proc_ghostCRIDs_ = new PromTable<PromTable<int>*>(tsz);
	  }
	  // get sends from CR
	  PromTable<int> *list = proc_ghostCRIDs_->find(proc+1);
	  if( list == NULL ) {
	    list = new PromTable<int>(tsz); CHKNEW(list,1);
	    proc_ghostCRIDs_->add( list, proc+1 );
	  }
	  list->add( xx + 1 ); // local nodes that touch an L.M.
	}
	else if( !constndf ) { // send 'ndf'
	  if( (table=proc_bounds.find(proc+1)) == NULL ) {
	    table = new PromTable<int>(tsz); CHKNEW(table,1);
	    proc_bounds.add( table, proc+1 );
	  }
	  table->add(xx+1); // send this to proc
	}
      }
    }
  }

  // form 'gid_ghostLid' 'proc_bounds' 
  ghostlid = nLocNd;
  if( haveelems ) {
    // construct promethues - extend 'elemList_' (needed under FEI)
    ierr = ExtendElemList(); CHKERRQ(ierr);    
    // form 'gid_ghostLid' with 'elemList_'
    Position pos = elemList_->getHeadPosition(); 
    while( pos ) { 
      int *arr = elemList_->getNext( pos );     
      for( ii = 0 ; ii < PROM_NENMAX ; ii++ ) { 
	gid = arr[ii]; 
	if( gid < 0 ) break;             // sentila or low rank elem
	if( gid < my0 || gid >= next0 ) {
	  if( (lid=gid_ghostLid.find(gid+1)-1) < 0 ) {
	    int add = gid_ghostLid.add(ghostlid+1,gid+1); assert(add==1);
	    lid = ghostlid++;
	  }
	}
      }
    }
  }
  else {
    // make fake elements
    assert(elemList_ == NULL); elemList_ = new PromList<int*>(1024);
    //
    int gids[4096]; // element ids
    for( xx = 0 ; xx < nLocNd ; xx++ ) {
      int *indices=blkColIndices[xx], len=blkRowLengths[xx]; assert(len<4096);
      for( jj = tt = 0 ; jj < len ; jj++ ) {
	int feigid = indices[jj], proc; bool lamb;
	ierr = id_map_private( feigid, &gid, &lamb, &proc ); CHKERRQ(ierr);
	if( !lamb ) {
	  gids[tt++] = gid;
	  if( proc != mype_ ) {
	    assert(npe_>1);
	    if( (lid=gid_ghostLid.find(gid+1)) == 0 ) {
	      int add = gid_ghostLid.add(ghostlid+1,gid+1); assert(add==1);
	      lid = ghostlid++;
	    }
	  }
	}
      }
      // make fake elems 
      for( jj = 0 ; jj < tt ; jj++ ) {
	int *arr, gidj = gids[jj], lidj = gidj - my0;
	if( lidj < 0 || lidj >= nLocNd || lidj > xx ) {
	  ierr = PetscMalloc(3*sizeof(int),&arr); CHKERRQ(ierr);
	  elemList_->addTail( arr ); // [ n1, n2, ... , nn, -matid ]
	  arr[0] = my0 + xx; arr[1] = gidj; arr[2] = -1;
	}
      }
    }
  }

  const int npGst = gid_ghostLid.getCount(), numnd = nLocNd + npGst;
  // send(tag) local node data - form 'lid_primLEq_' w/ 'proc_bounds'
  while( !proc_bounds.isEmpty() ) { 
    table = proc_bounds.removeHead( &proc ); proc--;
    // count data
    tt = 2*table->getCount(); // could be 0 (?)
    MPI_Request *sreq;
    ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq);  CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff; 
    // sends data
    while( !table->isEmpty() ) { 
      lid = table->removeHead()-1; gid = my0+lid; assert(lid>=0 && lid<nLocNd);
      *pb++ = gid; *pb++ = ptRowsPerBlkRow[lid];
    }
    if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
    // send(tag) *< gid, ndf >
    MPI_Isend( buff, tt, MPI_INT, proc, tag, wcomm, sreq );
    delete table; 
  }
  ierr = PetscMalloc((numnd+1)*sizeof(int),&lid_primLEq_); CHKERRQ(ierr);
  if( constndf ) {
    for( ii = 0 ; ii < numnd ; ii++ ) lid_primLEq_[ii] = ndf; 
    lid_primLEq_[numnd] = -1; // debug 
  }
  else {
    for(ii = 0 ; ii < nLocNd ; ii++) lid_primLEq_[ii] = ptRowsPerBlkRow[ii];
    for( /*VOID*/ ; ii < numnd+1 ; ii++ ) lid_primLEq_[ii] = -1; 
    // recv(tag) gid, ndf
    int ntodo = npGst;
    while( ntodo ){
      assert(ntodo > 0);
      MPI_Probe( MPI_ANY_SOURCE, tag, wcomm, &status );
      MPI_Get_count( &status, MPI_INT, &tt ); 
      assert( tt > 0 );
      int *buff,*pb;
      ierr = PetscMalloc(tt*sizeof(int),&buff); CHKERRQ(ierr); pb = buff;
      // recv(tag) *< gid, ndf >
      MPI_Recv(buff, tt, MPI_INT, status.MPI_SOURCE, tag, wcomm, &status);
      kk = tt/2; assert(tt%2 == 0);
   while( kk-- ) {
	ntodo--;
	gid = *pb++; int lndf = *pb++; 
	lid = gid_ghostLid.find(gid+1) - 1; assert(lid>=nLocNd && lid<numnd);
	lid_primLEq_[lid] = lndf; 
      }
      if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
      PetscFree(buff); 
    }
    while( !smessages.isEmpty() ) {
      MPI_Request *sreq = smessages.removeHead();
      MPI_Wait( sreq, &status );    PetscFree( sreq );
    }
    MPI_Iprobe( MPI_ANY_SOURCE, tag, wcomm, &tt, &status );   assert(!tt);
  }

  // core work
  ierr = setMatrixStructure_private( gid_ghostLid, ndf, mat );   CHKERRQ(ierr);
  prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FEI_INIT_STR]);
  
  PetscFunctionReturn(0);
}

/* Prometheus_LinSysCore::setMatrixStructure() ***************************
 *
 *  - Olympus initializer
 *
 *  INPUT: 
 *   - gid_ghostLid:
 *   - ndf: constant dof size
 *
 *  SIDE EFFECTS:
 *   - creates, and adds into, 'proc_ghostCRIDs_'
 */ 
int Prometheus_LinSysCore::setMatrixStructure(const PromTable<int> &gid_ghostLid,
					      const int ndf)
{
  int ierr, nLocNd = proc_gnode_[mype_+1] - proc_gnode_[mype_];
  PRINT;

  // reality check
  PetscFunctionBegin;
  if( proc_globalEq_ == NULL ){SETERRQ(1,"proc_globalEq_ == NULL");}  
  if( proc_globalEq_[npe_] != numGlobalRows_ ){SETERRQ(1,"ng!=pg[np]");}  
  if( PROM_BSZ < npe_ ){SETERRQ(1,"PROM_BSZ < np");}  

  prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FEI_INIT_STR]);
  // construct 'proc_ghostCRIDs_' with ;CR_id_;
  ierr = GetBCSends(); CHKERRQ(ierr);

  // core work
  const int numnd = nLocNd+gid_ghostLid.getCount(); assert(lid_primLEq_==0);
  ierr = PetscMalloc((numnd+1)*sizeof(int),&lid_primLEq_); CHKERRQ(ierr);
  for( int ii = 0 ; ii < numnd ; ii++ ) lid_primLEq_[ii] = ndf; 
  lid_primLEq_[numnd] = -1; // debug
  ierr = setMatrixStructure_private( gid_ghostLid, ndf ); CHKERRQ(ierr);
  prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FEI_INIT_STR]);
  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setMatrixStructure_private"
/* Prometheus_LinSysCore::setMatrixStructure_private() ******************
 *
 *  - construct core of Prometheus (called once only)
 *
 *  INPUT: 
 *   - gid_ghostLid
 *   - ndf: can be zero for VBR matrices
 *   - mat: PETSc Mat object - fine grid matrix (used in PETSc PC interface)
 *
 *  SIDE EFFECTS:
 */ 
int Prometheus_LinSysCore::setMatrixStructure_private(const PromTable<int> &gid_ghostLid,
						      const int ndf, void *mat /*= NULL*/ )
{
  // make Prometheus's fine grid -- " FEI: init. str."
  // create prometheus
  int ierr,ii;
  ierr = constructPrometheus( gid_ghostLid, ndf ); 
  CHKERRQ( ierr );
  PromGrid *grid0 = prom_->grid0_; assert(grid0 != NULL);
  const PromMap *const lmap = lmap_; 

  // matrices
  const bool have_const = (lmap != NULL), useCR = (have_const && !use_uzawa_);
  if( !have_const || !use_uzawa_ ){
    ierr = grid0->MakeMatrix( use_complex_, useCR, 0, NULL, mat );CHKERRQ(ierr); 
  }
  else { 
    assert( !use_complex_ ); // not supported
    ierr = grid0->MakeMatrix( use_complex_, useCR, 0, CR_id_, mat ); CHKERRQ(ierr);
  }
  PromMatrix *KK = grid0->stiffness_; 
  if( mat != NULL ) KK->malled_ = false; // hack for a user matrix!
  ierr = KK->SetOption( MAT_KEEP_ZEROED_ROWS );        CHKERRQ(ierr); 
  ierr = KK->SetOption( MAT_STRUCTURALLY_SYMMETRIC );  CHKERRQ(ierr); 

  if( have_const ){
    // constraint matrix
    if( use_uzawa_ ) {
      // normilization vector (diagonal matrix)
      ierr = PromUzawaSolver::Init( *grid0, *lmap_ ); CHKERRQ(ierr);
    }
    else {
      PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK);
      assert(K != NULL); assert(K->C_==NULL);
      K->C_ = C_; assert(C_!=NULL); C_ = NULL; // give Matrix projection CRs 
      grid0->CRmap_ = lmap_;            // give to fine grid
      // C'
      assert(grid0->scalarMap_== NULL);  assert(Ctnnz_!=NULL);
      grid0->scalarMap_ = new PromMap( grid0->nlocalEq(), 1, prom_->Comm_ );
      PromMatrix *CC = new PromMatrix(*grid0->scalarMap_, 0, Ctnnz_, 6, 
				      NULL,lmap_->nlocalEq(),numGlobalRowsLm_);
      if(Ctnnz_!=NULL) PetscFree(Ctnnz_); Ctnnz_ = NULL; assert(K->Ct_==NULL);
      K->Ct_ = CC;
    }
  }

  if( grid0->isConstNDF() && (!have_const || !use_uzawa_) )  {
    // set options after all is done -- if C than regularization adds
    ierr = KK->SetOption( MAT_NEW_NONZERO_ALLOCATION_ERR );  CHKERRQ(ierr);
  }

  // vectors -- b array
  assert(bVecArr_!=NULL); assert(numRHSs_>0);
  if( currentRHS_ < 0 ) currentRHS_ = 0;
  if( numRHSs_ < 0 ) { SETERRQ(1, "numRHSs_ < 0" ); }
  // Vecs, RHSs
  for( ii = 0; ii < numRHSs_; ii++ ){
    if( bVecArr_[ii] == NULL ){
      if( have_const ) bVecArr_[ii] = new PromCRVector( *grid0, lmap_ );
      else if( use_complex_ ) bVecArr_[ii] = new PromComplexVector( *grid0 ); 
      else bVecArr_[ii] = new PromVector( *grid0 ); 
      ierr = bVecArr_[ii]->Set( 0.0 );  CHKERRQ(ierr);  
    }
  }
  // init guess
  assert(init_guess_==NULL);
  if( have_const ) init_guess_ = new PromCRVector( *grid0, lmap ); 
  else if( use_complex_ ) init_guess_ = new PromComplexVector( *grid0 ); 
  else init_guess_ = new PromVector( *grid0 );
  ierr = init_guess_->Set(0.0); CHKERRQ(ierr);

  // xVec_ 
  assert(xVec_==NULL);
  if( have_const ) xVec_ = new PromCRVector( *grid0, lmap ); 
  else if( use_complex_ ) xVec_ = new PromComplexVector( *grid0 ); 
  else xVec_ = new PromVector( *grid0 );
  ierr = xVec_->Set(0.0); CHKERRQ(ierr);
  
  // work
  assert(work_==NULL);
  if( have_const ) work_ = new PromCRVector( *grid0, lmap ); 
  else if( use_complex_ ) work_ = new PromComplexVector( *grid0 ); 
  else work_ = new PromVector( *grid0 );
  ierr = work_->Set(0.0); CHKERRQ(ierr);
  
  // gamma_alpha
  assert(gamma_alpha_==NULL);
  if( have_const ) gamma_alpha_ = new PromCRVector( *grid0, lmap ); 
  else if( use_complex_ ) gamma_alpha_ = new PromComplexVector( *grid0 ); 
  else gamma_alpha_ = new PromVector( *grid0 );
  ierr = gamma_alpha_->Set( 0.0 ); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

/* Prometheus_LinSysCore::constructPrometheus() ***************************
 *
 *  -  construction
 *
 *  INPUT: 
 *   - gid_ghostLid: ghost ids
 *   - ndf: const ndf per node (0 if not const)
 *
 *  SIDE EFFECTS:
 *   - clears 'elemList_'
 */
int Prometheus_LinSysCore::constructPrometheus(const PromTable<int> &gid_ghostLid,
					       const int ndf )
{
  int ierr,ii,tt,jj;
  const int my0=proc_gnode_[mype_], next0=proc_gnode_[mype_+1];
  const int nLocNd = next0 - my0, nnds = nLocNd + gid_ghostLid.getCount(); 

  // new PromGrid
  ierr = prom_->CreateFineGrid( nLocNd, (ndf>0) ? NULL : lid_primLEq_,
				ndf, gid_ghostLid, elemList_ );
  CHKERRQ(ierr);

  // convert 'lid_primLEq_' to offset form
  for( tt = lid_primLEq_[0], ii = lid_primLEq_[0] = 0 ; ii < nnds ; ii++ ){ 
    jj = lid_primLEq_[ii+1]; lid_primLEq_[ii+1] = tt + lid_primLEq_[ii]; 
    tt = jj;
  } 

  // clear elemList_
  if( elemList_ != NULL ) {
    while( !elemList_->isEmpty() ){
      int *b = elemList_->removeHead();
      PetscFree( b );
    }
    delete elemList_; elemList_ = NULL;
  }

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::ExtendElemList"
/* Prometheus_LinSysCore::ExtendElemList() ***************************
 *
 *  - uses non-overlapping 'elemList_' to construct
 *
 *  INPUT: 
 *   - 
 *
 *  SIDE EFFECTS:
 *   - extents 'elemList_'
 */
int Prometheus_LinSysCore::ExtendElemList()
{
  int nen,tt,ii,jj,ierr,proc,gid,*arr;
  const int my0=proc_gnode_[mype_], next0=proc_gnode_[mype_+1];
  Position pos, pos2; const PromComm &comm = prom_->Comm_;
  const int tag = comm.getNewTag();
  PromTable<PromList<int*>*> proc_elemList(35);
  PromList<PromList<int*>*>procsElArrs(10); // temp list for 'proc_elemList'
  PromList<int*> *elList, deadElList;
  const int reqsz = sizeof(MPI_Request)/sizeof(int);
  PromList<MPI_Request*> smessages, rmessages; 
  MPI_Comm wcomm = MPIComm_;
  MPI_Status status;
  PRINT;
  
  if(elemList_ == NULL) elemList_ = new PromList<int*>(128);
  pos = elemList_->getHeadPosition(); nen = 0; 
  while( (pos2=pos) ){ 
    arr = elemList_->getNext(pos);  // [ n1, n2, ... , nn, -matid ]
    // see if pure ghost, copy locals gids
    for( ii = 0, jj = 0 ; arr[ii] >= 0 ; ii++ ) {
      gid = arr[ii];
      if(gid >= my0 && gid < next0) jj++; // count locals
    }
    if( ii > nen ) nen = ii; // get max nen
    
    if( jj == 0 ) { // pure foriegn
      elemList_->removeAt( pos2 ); // clip
      deadElList.addTail( arr ); // delete latter as needs to be sent
    }
    
    if(jj>1024)SETERRQ2(1,"buffer too small %d < %d",1024,jj); 
    if(!procsElArrs.isEmpty())SETERRQ(1,"!procsElArrs.isEmpty()");
    // ghosts: add element (arr), and local nodes, to send to ghost procs 
    for( ii = 0 ; arr[ii] >= 0 ; ii++ ) {
      gid = arr[ii];
      if( gid < my0 || gid >= next0 ) {
	// ghost's proc
	ierr = id_proc_private( gid, proc ); CHKERRQ(ierr); 
	if(proc==mype_)SETERRQ(1,"proc==mype");
	// add to 'proc_bounds' and 'proc_elemList'
	if( (elList=proc_elemList.find(proc+1)) == NULL ){
	  elList = new PromList<int*>(64); CHKNEW(elList,1);
	  proc_elemList.add(elList,proc+1);
	}
	// procs for this element -- small linear search in 'find'
	if( procsElArrs.find( elList ) == NULL ){
	  procsElArrs.addTail( elList );
	}
      }
    }
    // procs to send bounds (sbuf[jj])
    while( !procsElArrs.isEmpty() ) {
      elList = procsElArrs.removeHead();
      elList->addTail( arr ); // send element latter
    }
  }
  
  // count sends
  int rbuf[PROM_BSZ], sbuf[PROM_BSZ];
  for( ii = 0 ; ii < npe_ ; ii++ ) sbuf[ii] = 0;
  TablePos tpos = proc_elemList.getHeadPosition(); 
  while( tpos ) {
    elList = proc_elemList.getNext( tpos, &proc ); proc--; 
    tt = (nen+1)*elList->getCount() + 1;
    if(sbuf[proc]!=0)SETERRQ2(1,"sbuf[%d] = %d (!=0)",proc,sbuf[proc]);
    sbuf[proc] = tt;
  }
  // global!!
  MPI_Alltoall(sbuf,1,MPI_INT,rbuf,1,MPI_INT,wcomm);
  // post recv(tag)
  for( ii = 0 ; ii < npe_ ; ii++ ) {
    if( rbuf[ii] != 0 ) {
      // recv(tag)
      int sz = rbuf[ii], part = ii, *pb;
      ierr = PetscMalloc((sz+reqsz)*sizeof(int),&pb); CHKERRQ(ierr);
      MPI_Request *rreq = (MPI_Request*)pb; pb += reqsz;
      // irecv(tag) *< oldgid, loclid >
      MPI_Irecv( pb, sz, MPI_INT, part, tag, wcomm, rreq );
      rmessages.addTail( rreq );
    }
  }
  // send(tag) proc_elemList to other procs, count procs to do
  while( !proc_elemList.isEmpty() ) { 
    elList = proc_elemList.removeHead( &proc ); proc--;
    // count data
    tt = (nen+1)*elList->getCount() + 1;
    MPI_Request *sreq;
    ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq);  CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    // sends data
    *pb++ = nen; // dumb padding
    while( !elList->isEmpty() ) { 
      arr = elList->removeHead(); // arr is not freed here
      for( ii = 0, jj = 0 ; ii <= nen ; ii++ ) {
	if( jj == 0 ){ *pb++ = arr[ii]; if( arr[ii] < 0 ) jj = 1; }
	else *pb++ = -999; // dumb padding
      }
    }
    if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
    // send(tag) <nen, *< *<gid>, -matid [, -999, ..., -999] >>
    MPI_Isend( buff, tt, MPI_INT, proc, tag, wcomm, sreq );
    delete elList;  
  }
  // recv(tag) elements: add to 'elemList_'
  while( !rmessages.isEmpty() ) {
    MPI_Request *req = rmessages.removeHead(); 
    if(*req == MPI_REQUEST_NULL)SETERRQ(1,"*req == MPI_REQUEST_NULL");
    // recv(tag) < nen, *< *<gid>, -matid [, -999, ..., -9999] >>
    MPI_Wait( req, &status ); // determinate
    int *buff = (int*)req + reqsz, *pb = buff;   
    MPI_Get_count( &status, MPI_INT, &tt );
    nen = *pb++;
    while( pb-buff < tt ) {
      int *pp;
      ierr = PetscMalloc((nen+1)*sizeof(int),&pp); CHKERRQ(ierr);
      elemList_->addTail( pp ); // extend               
      for( ii = 0, jj = 0 ; /* VOID */ ; ii++, pp++ ) {
	gid = *pp = pb[ii]; // [ n1, n2, ... , nn, -matid [, junk ...]] 
	if( gid < 0 ) break;  // sentinal 
      }
      if( verbose_ > 4 || (mype_==0 && verbose_ > 3) ) {
	PromContext::printArr(stderr,pb,nen,mype_,"extra elem[%d]mype=%d");
      }
      pb += (nen+1);
    }
    if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
    PetscFree(req); 
  }
  // clean isend messages
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  while( !deadElList.isEmpty() ) { // free rest of elems
    arr = deadElList.removeHead(); PetscFree(arr);
  }

  return 0;
}

/************************************************************************
 * Matrix access functions
 ************************************************************************/
#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::putIntoSystemMatrix(point)"
/* Prometheus_LinSysCore::putIntoSystemMatrix() *******************
 *
 *  - point assembly 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::putIntoSystemMatrix( int numPtRows, 
						const int* ptRows,
						int numPtCols, 
						const int* ptCols,
						const double* const* values)
{
  return sumIntoSystemMatrix_private( numPtRows,ptRows,numPtCols,ptCols,values,
				      INSERT_VALUES, FALSE );
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::sumIntoSystemMatrix(point)"
/* Prometheus_LinSysCore::sumIntoSystemMatrix() *******************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumIntoSystemMatrix(int numPtRows, 
					       const int* ptRows,
					       int numPtCols, 
					       const int* ptCols,
					       const double*const* values)
{
  return sumIntoSystemMatrix_private(numPtRows,ptRows,numPtCols,ptCols,values,
				     ADD_VALUES, FALSE );
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::sumIntoSystemMatrix_i(point)"
/* Prometheus_LinSysCore::sumIntoSystemMatrix_i() *******************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumIntoSystemMatrix_i(int numPtRows, 
						 const int* ptRows,
						 int numPtCols, 
						 const int* ptCols,
						 const double*const* values)
{
  return sumIntoSystemMatrix_private(numPtRows,ptRows,numPtCols,ptCols,values,
				     ADD_VALUES, TRUE);
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::sumIntoSystemMatrix_private(point)"
/* Prometheus_LinSysCore::sumIntoSystemMatrix_private() *******************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumIntoSystemMatrix_private(int numPtRows, 
						       const int* ptRows,
						       int numPtCols, 
						       const int* ptCols,
						       const double*const*vals,
						       const int add_type,
						       const bool image )
{
  int ii,jj,ierr,geq,myeq,proc,nc,nlr,npr; bool lamb,projCR;
  PromMatrix *KK = prom_->grid0_->stiffness_; assert(KK!=NULL);
  int idsI[256],idsIlamb[256],idsJ[256],maskJ[256]; 
  double *pv,*pvlamb, vbuff[256*256], vbufflamp[256*256];

  if( numPtRows == 0 ) return 0;
  if( numPtRows > 256 || numPtCols > 256 ){
    SETERRQ2(1,"numPtRows=%d or numPtCols=%d > 256",numPtRows,numPtCols); 
  }
  if(verbose_> 4 || (mype_==0 && verbose_>3) ){
    PRINT4(numPtRows,*ptRows,numPtCols,*ptCols);
  }
  { static unsigned int cc = 0;
  if(verbose_>1 && (cc++)%1000 == 0)PetscPrintf(PETSC_COMM_WORLD,".");
  }
#if defined(PETSC_USE_BOPT_g)
  // " FEI: pt. assem."
  prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FEI_ASSEM]); 
#endif
  // cols, make mask
  for( jj = nc = 0 ; jj < numPtCols ; jj++ ){
    geq = ptCols[jj];
    ierr = eq_map_private( geq, &myeq, &lamb, &proc ); CHKERRQ(ierr);
    if( lamb ) maskJ[jj] = 0; // ignore C' or D (Alan)
    else{ maskJ[jj] = 1; idsJ[nc++] = myeq; } // A 
  }
  // rows and vals
  for(ii=nlr=npr=0,pv=vbuff,pvlamb=vbufflamp,projCR=TRUE;ii<numPtRows;ii++){
    geq = ptRows[ii];
    ierr = eq_map_private( geq, &myeq, &lamb, &proc ); CHKERRQ(ierr);
#if !defined(SALINAS_DD_SOLVER)
    if( proc != mype_ ){ SETERRQ(1,"row proc != myproc_");}
#endif
    if( lamb ) {
      idsIlamb[nlr++] = myeq; 
      if( projCR_id_ != NULL && projCR_id_->find(myeq+1) ) assert(projCR);
      else projCR = FALSE; 
    }
    else idsI[npr++] = myeq; 
 
    const double *varr = vals[ii]; 
    for( jj = 0 ; jj < numPtCols ; jj++ ) {
      if( maskJ[jj] == 1 ) {
	if( lamb ) *pvlamb++ = varr[jj]; 
	else *pv++ = varr[jj];
      }
    }
  }

  if( nc > 0 ) { 
    if(verbose_> 4 || (mype_==0 && verbose_>3) ){ PRINT3(npr,nlr,nc); }
    if( npr > 0 ) { // A 
      if( !image ) {
	ierr = KK->SetValues(npr,idsI,nc,idsJ,vbuff,add_type); CHKERRQ(ierr);
      }
      else {
	PromComplexMatrix *K = dynamic_cast<PromComplexMatrix*>(KK); 
	assert(K!=NULL);
	ierr = K->SetValues_i(npr,idsI,nc,idsJ,vbuff,add_type);  CHKERRQ(ierr);
      }
    }
    if( nlr > 0 ) {
      // cache v
      for( jj = 0 ; jj < nlr ; jj++ ) {
	PromCRCache *crc;
	crc = new PromCRCache( idsIlamb[jj], nc, idsJ, &vbufflamp[jj*nc] );
	if( projCR == FALSE ) {
	  crc->next_ = CRCache_; CRCache_ = crc; // regular CR
	}
	else {
	  crc->next_ = projCRCache_; projCRCache_ = crc; // projection CR
	}
      }
    }
  }
#if defined(PETSC_USE_BOPT_g)
  prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FEI_ASSEM]);
#endif
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::sumIntoSystemMatrix_i(blocked)"
/* Prometheus_LinSysCore::sumIntoSystemMatrix_i() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumIntoSystemMatrix_i(int numPtRows,const int*ptRows,
						 int numPtCols,const int*ptCols,
						 int numBlkRows,const int*blkRows,
						 int numBlkCols,const int*blkCols,
						 const double* const* values)
{
  return sumIntoSystemMatrix_private(numPtRows,ptRows,numPtCols,ptCols,
				     numBlkRows,blkRows,numBlkCols,
				     blkCols, values, TRUE ) ;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::sumIntoSystemMatrix(blocked)"
/*Prometheus_LinSysCore::sumIntoSystemMatrix() ***************************
 *
 *  - blocked assembly
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumIntoSystemMatrix(int numPtRows,
					       const int* ptRows,
					       int numPtCols,
					       const int* ptCols,
					       int numBlkRows,
					       const int* blkRows,
					       int numBlkCols,
					       const int* blkCols,
					       const double* const* values ) 
{
  return sumIntoSystemMatrix_private(numPtRows,ptRows,numPtCols,ptCols,
				     numBlkRows,blkRows,numBlkCols,
				     blkCols, values, FALSE ) ;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::sumIntoSystemMatrix_private(blocked)"
/*Prometheus_LinSysCore::sumIntoSystemMatrix_private() *******************
 *
 *  - blocked assembly
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumIntoSystemMatrix_private(int numPtRows,
						       const int* ptRows,
						       int numPtCols,
						       const int* ptCols,
						       int numBlkRows,
						       const int* blkRows,
						       int numBlkCols,
						       const int* blkCols,
						       const double*const*vals,
						       const bool image ) 
{
  int ii,jj,ierr,eqi,proc,mygid,feigid; bool lamb;
  PromMatrix *KK = prom_->grid0_->stiffness_;
  const int my0 = proc_gnode_[mype_], nloc=KK->map_.nlocal(); double *pv;
  int idsI[256],idsJ[256]; double fbuff[256*256]; 

  if( numBlkRows > 256 || numBlkCols > 256 ){
    SETERRQ2(1,"numBlkRows=%d or numBlkCols=%d > 256",numBlkRows,numBlkCols); 
  }
  else if( numBlkRows == 0 ) return 0;
  else  if(verbose_ > 4 || (mype_==0 && verbose_>3) ){
    PRINT2(numBlkRows,numBlkCols);
  }
  { static unsigned int cc = 0;
  if(verbose_>1 && (cc++)%5000 == 0)PetscPrintf(PETSC_COMM_WORLD,"*");
  }
#if defined(PETSC_USE_BOPT_g) 
  prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_METIS]);
#endif
  // cols
  for( jj = 0 ; jj < numBlkCols  ; jj++ ){
    feigid = blkCols[jj]; 
    ierr = id_map_private( feigid, &mygid, &lamb, &proc ); CHKERRQ(ierr);
    if( lamb ){ SETERRQ(1,"lamb"); } 
    idsJ[jj] = mygid;
  }
  // rows
  for( ii = eqi = 0, pv = fbuff ; ii < numBlkRows ; ii++ ){
    feigid = blkRows[ii];
    ierr = id_map_private( feigid, &mygid, &lamb, &proc ); CHKERRQ(ierr);
    if( lamb ){ SETERRQ(1,"lamb"); } 
    idsI[ii] = mygid;
    // vals
    int lid = mygid-my0, ndf = lid_primLEq_[lid+1]-lid_primLEq_[lid];
    assert(lid>=0 && lid<nloc && ndf>0);
    for( int xx = 0 ; xx < ndf ; xx++, eqi++ ) {
      const double *varr = vals[eqi];
      for(jj = 0 ; jj < numPtCols ; jj++) *pv++ = varr[jj];
    }
  }
  if( eqi != numPtRows ) SETERRQ2(1,"eqi != numPtRows %d %d ",eqi,numPtRows);
  //
  if( !image ) {
    ierr = KK->SetValuesBlocked( numBlkRows, idsI, numBlkCols, idsJ, fbuff,
				 ADD_VALUES);
    if(ierr!=0)PromContext::printArr(stderr,idsJ,numBlkCols,"idsJ[%d]");
    CHKERRQ(ierr);
  }
  else {
    PromComplexMatrix *K = dynamic_cast<PromComplexMatrix*>(KK);
    assert(K!=NULL);
    ierr = K->SetValuesBlocked_i( numBlkRows, idsI, numBlkCols, idsJ, fbuff,
				  ADD_VALUES );
    CHKERRQ(ierr);
  }

#if defined(PETSC_USE_BOPT_g)
  prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_METIS]);
#endif  
  return 0;
}

/************************************************************************
 * Vector access functions
 ************************************************************************/

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::putInitialGuess"
/* Prometheus_LinSysCore::putInitialGuess() ***************************
 *
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::putInitialGuess( const int* eqnNumbers,
					    const double* values, 
					    int len) 
{
  //
  //This function scatters (puts) values into the linear-system's soln vector.
  //
  // num is how many values are being passed,
  // indices holds the global 'row-numbers' into which the values go,
  // and values holds the actual coefficients to be scattered.
  //
  if(verbose_>2) PRINT1(len);
  
  if( len ) {
    int ii,jj,geq,myeq,lamb1=99,ierr,proc; bool lamb;
    int *rows = NULL; double *vv = NULL; 
    int rbuf[PROM_BSZ]; double fbuf[PROM_BSZ]; 
    if( len > PROM_BSZ ) {
      ierr = PetscMalloc(len*sizeof(int),&rows); CHKERRQ(ierr);
      ierr = PetscMalloc(len*sizeof(double),&vv); CHKERRQ(ierr);
    }
    else { vv = fbuf; rows = rbuf; }
   
    //
    for( ii = jj = 0 ; ii < len ; ii++ ){
      geq = eqnNumbers[ii];
      ierr = eq_map_private( geq, &myeq, &lamb, &proc ); CHKERRQ(ierr);
      if( proc == mype_ && values[ii] != 0.0 ){
	vv[jj] = values[ii]; rows[jj] = myeq; jj++;
	if( lamb1 == 99 ) lamb1 = lamb;
	else if( lamb1 != lamb ) { SETERRQ(1, "lamb1 != lamb" ); }
      }
    }
    if( jj != 0 ){  // local -- redundent guesses are provided
      if( lamb1 ) { 
	//init_guess_->p_->SetValues(jj,rows,vv,INSERT_VALUES);CHKERRQ(ierr);
	SETERRQ(1,"putInitialGuess in langrange mult.????" ); 
      }
      else {
	ierr = init_guess_->SetValues(jj,rows,vv,INSERT_VALUES); CHKERRQ(ierr);
	dirty_guess_ = TRUE;
      }
    }
    if( rows != rbuf ) { PetscFree(rows); PetscFree(vv); }
  }
  
  return 0;
}
 
#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::getFromRHSVector"
/* Prometheus_LinSysCore::getFromRHSVector() ***************************
 * 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getFromRHSVector( int num, double* values, 
					     const int* indices)
{
  // hack to accomodate getvalues
  if(verbose_>3)PRINT1(num);
  if( num == 0 ) return 0;
  
  int ierr,*rows,ii,_rbuf[PROM_BSZ];
  if( num < PROM_BSZ ) rows = _rbuf;
  else{ ierr = PetscMalloc(num*sizeof(int),&rows); CHKERRQ(ierr); }
  for(ii=0;ii<num;ii++) rows[ii] = indices[ii]; 
  //
  ierr = accessRHSVector_private( num, values, rows, PROM_GET_VALUES, FALSE); 
  CHKERRQ(ierr);

  if( rows != _rbuf ) PetscFree(rows); 
  
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::putIntoRHSVector"
/*
 * - 
 */
int Prometheus_LinSysCore::putIntoRHSVector( int num, const double* values,
					     const int* indices )
{
  // hack to accomodate getvalues
  if(INSERT_VALUES == PROM_GET_VALUES)SETERRQ(1,"INSERT == GET_VALUES");
  int ierr;

  ierr = sumIntoRHSVector_private(num, values, indices, INSERT_VALUES, FALSE);
  CHKERRQ(ierr);

  return 0;
}
  
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::sumIntoRHSVector"
/*
 * - 
 */
int Prometheus_LinSysCore::sumIntoRHSVector( int num, 
					     const double* values,
					     const int* indices ) 
{
  // hack to accomodate getvalues
  if(ADD_VALUES==PROM_GET_VALUES)SETERRQ(1,"ADD == GET_VALUES");
  int ierr;

  ierr = sumIntoRHSVector_private(num,values,indices,ADD_VALUES,FALSE);
  CHKERRQ(ierr);
  
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::sumIntoRHSVector_i"
/* Prometheus_LinSysCore::sumIntoRHSVector_i() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumIntoRHSVector_i( int num, const double* values, 
					       const int* indices )
{
  if(verbose_>2) PRINT1(num);
 
  int ierr = sumIntoRHSVector_private(num,values,indices,ADD_VALUES,TRUE);
  CHKERRQ(ierr);
  
  return 0;
}
 
#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::putIntoRHSVector_i"
/* Prometheus_LinSysCore::putIntoRHSVector_i() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::putIntoRHSVector_i( int num, const double* values, 
					       const int* indices )
{
  if(verbose_>2) PRINT1(num);
 
  return sumIntoRHSVector_private(num,values,indices,INSERT_VALUES,TRUE);
}
 
#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::getFromRHSVector_i"
/* Prometheus_LinSysCore::getFromRHSVector_i() ***************************
 * 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getFromRHSVector_i( int num, double* values, 
					       const int* indices )
{
  // hack to accomodate getvalues
  if(verbose_>3)PRINT1(num);
  if( num == 0 ) return 0;
  
  int ierr,*rows,ii,_rbuf[PROM_BSZ];
  if( num < PROM_BSZ ) rows = _rbuf;
  else{ ierr = PetscMalloc(num*sizeof(int),&rows); CHKERRQ(ierr); }
  for(ii=0;ii<num;ii++) rows[ii] = indices[ii]; 
  //
  ierr = accessRHSVector_private(num, values, rows, PROM_GET_VALUES, TRUE); 
  CHKERRQ(ierr);

  if( rows != _rbuf ) PetscFree(rows); 
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::sumIntoRHSVector_private"
/* Prometheus_LinSysCore::sumIntoRHSVector_private() *********************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumIntoRHSVector_private( const int num, 
						     const double* values,
						     const int* indices,
						     const int add_type,
						     const bool image ) 
{
  int ierr,ii,*rows; double *v1,*vv;
  double fbuf[PROM_BSZ]; int rbuf[PROM_BSZ];
  
  if(verbose_>4 || (mype_==0 && verbose_>3)) {
    PRINT2(num,add_type);
  }
  if( num == 0 ) return 0;
  
  if( num < PROM_BSZ ){
    v1 = vv = fbuf; rows = rbuf;
  }
  else{ 
    ierr = PetscMalloc(num*sizeof(double),&vv); CHKERRQ(ierr); v1 = vv;
    ierr = PetscMalloc(num*sizeof(int),&rows); CHKERRQ(ierr); 
  }
  for(ii=0;ii<num;ii++){ v1[ii] = values[ii]; rows[ii] = indices[ii]; }
  // 
  ierr = accessRHSVector_private( num, v1, rows, add_type, image ); 
  CHKERRQ(ierr);
  
  if( rows != rbuf ){ PetscFree(vv); PetscFree(rows); }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::accessRHSVector_private"
/* Prometheus_LinSysCore::accessRHSVector_private() *********************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::accessRHSVector_private(const int num, 
						   double* values,
						   int* indices,
						   const int add_type,
						   const bool image ) 
{
  //
  //This function scatters (accumulates) values into the linear-system's
  //currently selected RHS vector.
  //
  // num is how many values are being passed,
  // indices holds the global 'row-numbers' into which the values go,
  // and values holds the actual coefficients to be scattered.
  //
  if( num == 0 ) return 0;
  int ii,geq,myeq,lamb1=99,proc,ierr,nnz;
  bool lamb; double *pv;
  if(currentRHS_<0) SETERRQ(1, "currentRHS_ < 0" );
  PromVector *BB = bVecArr_[currentRHS_]; 
  if(BB == NULL) SETERRQ(1, "bVecArr_[currentRHS_] == NULL" );
  if( verbose_ > 4 || (mype_==0 && verbose_ > 3) ) PRINT1(num);
  const int fact = BB->isComplex() ? 2 : 1, ioffset = image ? 1 : 0;

  // map indices
  for( ii = nnz = 0 ; ii < num ; ii++ ){ 
    geq = indices[ii]; 
    if(values[ii]!=0.0||add_type==INSERT_VALUES || add_type==PROM_GET_VALUES){
      ierr = eq_map_private( geq, &myeq, &lamb, &proc ); CHKERRQ(ierr);
      if( lamb1 == 99 ) lamb1 = lamb; 
      else if(lamb1!=lamb||proc!=mype_){SETERRQ(1,"pm1!=pm || pe!=mype");}
      indices[nnz] = myeq; // overwrites 'indices' with ly indices
      values[nnz] = values[ii]; 
      nnz++;
    }
  }
  // doit
  if ( add_type == PROM_GET_VALUES ) { // hack to accomodate getvalues
    if( lamb1 ){ 
      const int lm0 = proc_lmEq_[mype_];
      PromCRVector *B = dynamic_cast<PromCRVector*>(BB); assert(B!=NULL);
      ierr = B->p_->GetArray( &pv );    CHKERRQ(ierr);
      for(int ii=0;ii<num;ii++) values[ii] = pv[indices[ii] - lm0];
      ierr = B->p_->RestoreArray( &pv );    CHKERRQ(ierr);
    }
    else{
      const int prim0 = proc_primEq_[mype_];
      ierr = BB->GetArray( &pv );    CHKERRQ(ierr);
      for(int ii=0;ii<num;ii++) {
	int lid = (indices[ii]-prim0)*fact + ioffset; assert(lid>=0);
	values[ii] = pv[lid];
      }
      ierr = BB->RestoreArray( &pv );    CHKERRQ(ierr);
    } 
  }
  else if( nnz > 0 ) {
    PromCRVector *B = dynamic_cast<PromCRVector*>(BB); assert(!lamb||B!=NULL);
    PromVector *vec = lamb1 ? B->p_ : BB;
    if(vec == NULL) SETERRQ1(1, "BB->p_==NULL, lamb1=%d",lamb1);
    if( INSERT_VALUES == add_type ) {
      const int my0 = lamb1 ? proc_lmEq_[mype_] : proc_primEq_[mype_];
      ierr = vec->GetArray( &pv );    CHKERRQ(ierr);
      for( ii = 0 ; ii < nnz ; ii++ ){
	int lid = (indices[ii]-my0)*fact + ioffset; assert(lid>=0);
	values[ii] = values[ii] - pv[lid];
      }
      ierr = vec->RestoreArray( &pv ); CHKERRQ(ierr); 
    }
    if( !image ) {
      ierr = vec->SetValues( nnz, indices, values, ADD_VALUES ); CHKERRQ(ierr);
    }
    else {
      PromComplexVector *cv = dynamic_cast<PromComplexVector*>(vec); 
      assert(cv!=NULL);
      ierr = cv->SetValues_i(nnz, indices, values, ADD_VALUES); CHKERRQ(ierr);
    }
  }
  return 0;
}

/************************************************************************
 * matrixLoadComplete -- set up for solve
 ************************************************************************/
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::matrixLoadComplete"
/* Prometheus_LinSysCore::matrixLoadComplete() ***************************
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *    - deletes 'CR_id_'s if 'use_uzawa_'
 *    - 'C_' has projection matrix
 */
int Prometheus_LinSysCore::matrixLoadComplete() 
{
  PromGrid *grid0 = prom_->grid0_;
  PromMatrix *KK = grid0->stiffness_; 
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); 
  PromMatrix *CC = use_uzawa_ ? uzawaC_ : (K==NULL ? NULL : K->C_);
  PromTable<int> CReq_projCReq(5);
  int *logs = PromContext::prom_logs_, matsetup = logs[LOG_FEI_MATRIX_SETUP];
  int bccr = logs[LOG_FEI_CR_BC], kkt=logs[LOG_AUX2], vb = verbose_;
  MPI_Comm wcomm = MPIComm_;
  PRINT;

  // detect dirty
  PetscFunctionBegin;
  int ierr,tt[2],cc = (CRCache_==NULL && projCRCache_==NULL) ? 0 : 1;
  int dd[2]={ (KK->isDirty() ? 1 : 0), cc };
  MPI_Allreduce(dd, tt, 2, MPI_INT, MPI_MAX, wcomm );
  KK->SetDirty( tt[0] != 0 ); 
  if( CC != NULL ) CC->SetDirty( tt[1] != 0 );
  else assert(tt[1] == 0);
  if( C_ != NULL ) C_->SetDirty( tt[1] != 0 );

  if( tt[1] != 0 && !KK->isDirty() ) {
    assert(0); // not used!!!
    assert(CC!=NULL); CC->SetDirty(); // we just have dirty C
    // make solver object -- " FEI: Prom setup"
    prom_->perf_mon_.EventBegin(matsetup);
    // get BC data -- " FEI: BCs & reg."
    prom_->perf_mon_.EventBegin(bccr);
    //get sends into 'proc_ghostCRIDs_'(1st pass done in 'constructPrometheus')
    ierr = GetBCSends( ); CHKERRQ(ierr);
    // need to update BCs -- empties 'proc_ghostCRIDs_'
    ierr = SetBCforCRTable( CC, C_ ); CHKERRQ(ierr);
    // assemble CR equations from cache
    ierr = assembleCRs( &CReq_projCReq ); CHKERRQ(ierr); 
    prom_->perf_mon_.EventEnd(bccr);

    // redo matrix setup
    ierr = ReFactorSolver();                 CHKERRQ(ierr);
    // now clean
    CC->SetDirty( FALSE );
    KK->SetDirty( FALSE );
    if( C_ != NULL ) C_->SetDirty( FALSE );    
    assert( !CC->isDirty() );
    assert(essbcdata_ == NULL); // no new BCs
    assert(othbcdata_ == NULL); // no new BCs
    assert( !KK->isDirty() ); 
    if( C_ != NULL ) {
      assert( use_uzawa_ );
      // " FEI: BCs & reg."
      prom_->perf_mon_.EventBegin(bccr);
      PromSolver *const solver = prom_->solver_; assert(solver!=NULL);
      assert(solver->projector_ && solver->projector_->projC_ == NULL );
      solver->projector_->SetProjC( C_, &CReq_projCReq );
      C_ = NULL; 
      prom_->perf_mon_.EventEnd(bccr);
    }
    prom_->perf_mon_.EventEnd(matsetup);
  }
  else if( KK->isDirty() ) {
    // make solver object -- " FEI: Prom setup"
    prom_->perf_mon_.EventBegin(matsetup);
    // apply BCs (before reg.) -- " FEI: BCs & reg."
    prom_->perf_mon_.EventBegin(bccr);
    // add dummy regularization, before assembly
    if( prom_->solver_ == NULL && use_uzawa_ && CC != NULL ) {
      if(vb>1)PetscPrintf(wcomm,"[%d]%s Set dummy reg\n",mype_,__FUNCT__);
      ierr = setDummyReg( KK, CR_id_ );  CHKERRQ(ierr);
      // delete 'CR_id_' (from fine grid) and save coarse grids
      while( !CR_id_->isEmpty() ){ 
	PromCR* mcr = CR_id_->removeHead(); delete mcr;
      }
    }
    else if( use_uzawa_ && CC != NULL ) {
      assert(0); // not used
      ierr = GetBCSends(); CHKERRQ(ierr);
    }

    if( prom_->solver_ == NULL ) {
      // assembles matrix sets node.fix(), creates zeroIS_, sets 'gamma_alpha_'
      if(vb>1)PetscPrintf(wcomm,"[%d]%s collect BC data\n",mype_,__FUNCT__);
      ierr = collectBCs(); CHKERRQ(ierr); 
      // communicate BC data
      if(vb>1)PetscPrintf(wcomm,"[%d]%s send BC data\n",mype_,__FUNCT__);
      // need to update BCs -- empties 'proc_ghostCRIDs_'
      ierr = SetBCforCRTable( CC, C_ );            CHKERRQ(ierr);
      // assemble CR equations from cache, after BCs
      ierr = assembleCRs( &CReq_projCReq ); CHKERRQ(ierr);
    }
    else {
      ierr = KK->Assembly();   CHKERRQ(ierr);
    }
    ierr = bVecArr_[currentRHS_]->Assembly();   CHKERRQ(ierr);
    prom_->perf_mon_.EventEnd(bccr);
    
    // apply BCs to AA
    ierr = applyBCs( C_ ); CHKERRQ(ierr);
    prom_->perf_mon_.EventEnd(matsetup);
    
    // doit it
    if( prom_->solver_ == NULL ) {
      // create grids and solver, scales A (uses 'CR_id_')
      ierr = SymbolicSetUp();                 CHKERRQ(ierr);
      
      // " FEI: solv.setup"
      prom_->perf_mon_.EventBegin(matsetup);
      ierr = MakeSolver( (K==NULL || K->C_==NULL) ? NULL : &K->C_->map_, 
			 !KK->isMalloced() );   CHKERRQ(ierr);
      PromSolver *const solver = prom_->solver_; assert(solver!=NULL);

      if( C_ != NULL ) { // cached projection matrix in 'C_'
	// " FEI: BCs & reg."
	prom_->perf_mon_.EventBegin(kkt); 
	assert(use_uzawa_);
	assert(solver->projector_ == NULL );
	solver->projector_ = new PromProjKKTSolver();
	ierr = solver->projector_->SetProjC( C_, &CReq_projCReq ); 
	CHKERRQ(ierr);
	C_ = NULL; 
	prom_->perf_mon_.EventEnd(kkt); 
      }

      // solver setup -- " FEI: solv.setup"
      ierr = solver->SetUp(); CHKERRQ(ierr);
      prom_->perf_mon_.EventEnd(matsetup);
    }
    else {
      // " FEI: solv.setup"
      prom_->perf_mon_.EventBegin(matsetup);
      
      PromSolver *const solver = prom_->solver_; assert(solver!=NULL);
      if( C_ != NULL ) {
	// " FEI: BCs & reg."
	prom_->perf_mon_.EventBegin(bccr); 
	assert(solver->projector_ && solver->projector_->projC_ == NULL );
	assert( use_uzawa_ );
	solver->projector_->SetProjC( C_, &CReq_projCReq);
	C_ = NULL; 
	prom_->perf_mon_.EventEnd(bccr); 
      }

      // refactor
      ierr = ReFactorSolver(); CHKERRQ(ierr); 
      prom_->perf_mon_.EventEnd(matsetup); 
    }
    KK->SetDirty( FALSE ); // this is now 'clean' -- set in a->setval (ugh!!)
  }

  // delete CR grid data
  if( CR_id_ != NULL ) {
    while( !CR_id_->isEmpty() ){ 
      PromCR* mcr = CR_id_->removeHead(); delete mcr;
    }
  }
  if( projCR_id_ != NULL ) {
    while( !projCR_id_->isEmpty() ){ 
      PromCR* mcr = projCR_id_->removeHead(); delete mcr;
    }
  }
  
  assert( CC == NULL || !CC->isDirty() );
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::GetBCSends"
/* Prometheus_LinSysCore::GetBCSends() ***************************
 *
 *  get BC data for new ghosts from CRs
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *   - fills 'proc_ghostCRIDs_' from 'CR_id_'
 */
int Prometheus_LinSysCore::GetBCSends() 
{
  if( numGlobalRowsLm_ == 0 ) return 0;
  // communicate BC data -- do not have it after init solve
  int ierr,proc,tabsz = (4*(proc_lmEq_[mype_+1]-proc_lmEq_[mype_])/3+5); 
  const int reqsz = sizeof(MPI_Request)/sizeof(int); 
  MPI_Comm wcomm = MPIComm_;
  PromList<MPI_Request*> smessages; 
  MPI_Status    status;  const int tag = prom_->Comm_.getNewTag();
  PromTable<PromTable<int>*> proc_list(57); PromTable<int> *list; 
  const int my0=proc_gnode_[mype_],next0=proc_gnode_[mype_+1];
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 

  // collect requests
  assert(proc_ghostCRIDs_ == NULL || proc_ghostCRIDs_->isEmpty());
  TablePos tpos = CR_id_->getHeadPosition();
  while( tpos ) {
    PromCR* mcr = CR_id_->getNext( tpos ); 
    int *pi = mcr->gnodes_, len = mcr->len(); 
    // send requests
    for( int ii = 0 ; ii < len ; ii++ ) {
      int gid = pi[ii];
      if( gid < my0 || gid >= next0 ) {
	PromGrid::FindProc(proc_gnode_,npe_,gid,proc); assert(proc!=mype_);
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromTable<int>(tabsz); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	list->add( gid+1 ); 
      }
    }
  }
  if( projCR_id_ != NULL ) {
    tpos = projCR_id_->getHeadPosition();
    while( tpos ) {
      PromCR* mcr = projCR_id_->getNext( tpos ); 
      int *pi = mcr->gnodes_, len = mcr->len(); 
      // send requests
      for( int ii = 0 ; ii < len ; ii++ ) {
	int gid = pi[ii];
	if( gid < my0 || gid >= next0 ) {
	  PromGrid::FindProc(proc_gnode_, npe_, gid, proc);assert(proc!=mype_);
	  if( (list=proc_list.find(proc+1)) == NULL ) {
	    list = new PromTable<int>(tabsz); CHKNEW(list,1);
	    proc_list.add( list, proc+1 );
	  }
	  list->add( gid+1 ); 
	}
      }
    }
  }
  // send(tag) *< gid >, need these gid's BCs
  for( int ii = 0 ; ii < npe_ ; ii++ ) sbuf[ii] = 0;  
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    int tt = list->getCount(); 
    MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int), &sreq);
    CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) *pb++ = list->removeHead() - 1;  
    if(pb-buff != tt)SETERRQ2(1,"pb-buff != tt; %d != %d",pb-buff,tt); 
    // send(tag) *< gid >
    MPI_Isend( buff, tt, MPI_INT, proc, tag, wcomm, sreq );
    delete list;
    sbuf[proc] = tt; assert(tt>0);
  }

  // recive requests -- global!
  MPI_Alltoall( sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, wcomm );
  for( proc = 0 ; proc < npe_ ; proc++ ) {
    if( rbuf[proc] != 0 ) {
      // recv(tag)
      int tt = rbuf[proc], *buff, *pb; 
      ierr = PetscMalloc(tt*sizeof(int),&buff );  CHKERRQ(ierr);
      pb = buff;
      // recv(tag) *< gid >
      MPI_Recv( buff, tt, MPI_INT, proc, tag, wcomm, &status );
      MPI_Get_count( &status, MPI_INT, &ierr ); assert(tt==ierr);
      proc = status.MPI_SOURCE;
      while( pb-buff < tt ) {
	int gid = *pb++; assert(gid >= my0 && gid < next0);
	if( proc_ghostCRIDs_ == NULL ) {
	  proc_ghostCRIDs_ = new PromTable<PromTable<int>*>(tabsz);
	}
	// get sends from CR
	if( (list=proc_ghostCRIDs_->find(proc+1)) == NULL ) {
	  list = new PromTable<int>(tabsz); CHKNEW(list,1);
	  proc_ghostCRIDs_->add( list, proc+1 );
	}
	list->add( gid - my0 + 1 ); 
      }
      PetscFree(buff); assert(pb-buff == tt);    
    }
  }
  // cleanup sends *< gid, par_gid >
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }
  MPI_Iprobe( MPI_ANY_SOURCE, tag, wcomm, &proc, &status ); assert(!proc);

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::SetBCforCRTable"
/* Prometheus_LinSysCore::SetBCforCRTable() ***************************
 *
 *   - receive needed gids (in 'geq_fixity_') BCs in 'proc_ghostCRIDs_'
 *
 *  SIDE EFFECTS:
 *   - puts ghost BCs in 'geq_fixity_'
 *   - empties 'proc_ghostCRIDs_'
 */
int Prometheus_LinSysCore::SetBCforCRTable(const PromMatrix * const CC, 
					   const PromMatrix *const projC)  
{ 
  // communicate BC data -- do not have it after init solve
  int ierr, proc,tt,todo,ii; 
  const int reqsz=sizeof(MPI_Request)/sizeof(int),my0eq = proc_primEq_[mype_];
  MPI_Comm wcomm = MPIComm_;
  PromList<MPI_Request*> smessages; const int tag = prom_->Comm_.getNewTag();
  const PromNode * const * nodes = prom_->grid0_->nodes_;
  MPI_Status status; TablePos tpos;
  const int tabsz = 4*(proc_lmEq_[mype_+1]-proc_lmEq_[mype_])/3 + 17;
  double *bcp, bc; float *fp; assert(sizeof(float) == sizeof(int));
  const int fact = gamma_alpha_->isComplex() ? 2 : 1;

  // send *<gid, nfixed, nfixed*<geq, bc>>
  ierr = gamma_alpha_->GetArray( &bcp );    CHKERRQ(ierr);
  while( proc_ghostCRIDs_ != NULL && !proc_ghostCRIDs_->isEmpty() ) {
    PromTable<int> *list = proc_ghostCRIDs_->removeHead(&proc); proc--;
    // count
    tpos = list->getHeadPosition(); tt = 0;
    while( tpos ){
      int lid = list->getNext( tpos ) - 1;
      const PromNode *curr = nodes[lid];
      int ndf = curr->getNDF(), nf = 0;
      for( int jj = 0 ; jj < ndf ; jj++ ) if( curr->isFixed(jj) ) nf++;
      tt += 2 + 2*nf;
    }
    MPI_Request *sreq; 
    ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    while( !list->isEmpty() ) {
      int lid = list->removeHead() - 1, *pb2; 
      const PromNode *curr = nodes[lid];
      *pb++ = curr->index_.global(); //*pb++ = curr->geq(); *pb++ = curr->getNDF();
      pb2 = pb; *pb++ = -2; 
      int ndf = curr->getNDF(), geq = curr->geq(), nf = 0;
      for( int jj = 0 ; jj < ndf ; jj++, geq++ ) {
	if( curr->isFixed( jj ) ) {
	  nf++; int leq = fact*(geq - my0eq); assert(leq>=0);
	  *pb++ = geq;
	  fp = (float*)pb; *fp++ = (float)bcp[leq]; pb = (int*)fp;
	}
      }
      *pb2 = nf;
    }
    assert(tt == pb-buff);
    // send(tag) *<gid, nfixed, nfixed*<geq, bc>>
    MPI_Isend( buff, pb-buff, MPI_INT, proc, tag, wcomm, sreq );
    delete list;
  } 
  delete proc_ghostCRIDs_; proc_ghostCRIDs_ = NULL;

  // cached needed ghost eqs in 'geq_fixity_'
  if( geq_fixity_ == NULL ) todo = 0;
  else{ todo = geq_fixity_->getCount(); geq_fixity_->removeAlld(); }
  // set local fixities for CRs
  if( CC || projC ) {
    PromIS zis = (PromIS)zeroIS_; assert( zis != NULL ); int *ids,sz;
    ierr = PromISGetIndices( zis, &ids ); CHKERRQ(ierr);
    ierr = PromISGetSize( zis, &sz ); CHKERRQ(ierr);
    if( verbose_ > 2 ) PromContext::printArr(stderr,ids,sz,"zeroBCs[%d]");
    if(sz>0 && geq_fixity_==NULL)geq_fixity_=new PromTable<double>(tabsz);
    for( ii = 0 ; ii < sz ; ii++ ) {
      int geq = ids[ii], leq = geq - my0eq; bc = bcp[leq];
      int add = geq_fixity_->addd( bc, geq + 1 ); assert(add); 
    }
    ierr = PromISRestoreIndices( zis, &ids ); CHKERRQ(ierr);
  }
  ierr = gamma_alpha_->RestoreArray( &bcp );    CHKERRQ(ierr);

  //recv(tag)
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, wcomm, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); proc = status.MPI_SOURCE;
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // recv(tag) *<gid, nfixed, nfixed*<geq, bc>>
    MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, wcomm, &status );
    while( tt > pb-buff ) {
      todo--; assert(todo>=0);
      ii = *pb++; // gid
      int nf = *pb++; 
      for( int jj = 0 ; jj < nf ; jj++ ){
	int geq = *pb++; 
	fp = (float*)pb; bc = (double)*fp++; pb = (int*)fp;
	int add = geq_fixity_->addd( bc, geq + 1 ); assert(add); 
      }
    }
    assert(tt == pb-buff);
    PetscFree( buff );
  }
  // cleanup sends
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  } 
  // 
  MPI_Iprobe( MPI_ANY_SOURCE, tag, wcomm, &ii, &status ); assert(!ii);

  return 0;
}

/************************************************************************
 * BC functions
 ************************************************************************/

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::enforceEssentialBC"
/* Prometheus_LinSysCore::enforceEssentialBC() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::enforceEssentialBC( int* globalEqn, double* alpha,
					       double* gamma, int len) 
{
  if( len ) {
    PRINT2(len,*globalEqn); 
    Prometheus_EssBCData *bc = 
      new Prometheus_EssBCData( globalEqn, alpha, gamma, len );
    CHKNEW(bc,1);
    bc->next_ = essbcdata_;
    essbcdata_ = bc;
  }
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::enforceOtherBC"
/* Prometheus_LinSysCore::enforceOtherBC() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::enforceOtherBC( int* globalEqn, double* alpha,
					   double* beta, double* gamma, 
					   int len)
{
  if( len ) {
    PRINT2(len,*globalEqn); 
    Prometheus_OthBCData *bc =
      new Prometheus_OthBCData(globalEqn,alpha,beta,gamma,len);
    CHKNEW(bc,1);
    bc->next_ = othbcdata_;
    othbcdata_ = bc;
  }
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::enforceEssentialBC_private"
/* Prometheus_LinSysCore::enforceEssentialBC_private() ***********************
 *
 *  - set node fixity and create BC data 'zeroEqs_'
 *
 *  INPUT: 
 *   - globalEqn[len]:
 *   - alpha[len]: ** (in/out) ** returns term for RHS
 *   - gamma[len]:
 *   - len:
 *
 *  SIDE EFFECTS:
 *    - adds to 'zeroEqs'
 */
int Prometheus_LinSysCore::enforceEssentialBC_private(int* globalEqn,
						      double* alpha,
						      double* gamma, int len,
						      PromVector *bmod ) 
{
  //
  //This function must enforce an essential boundary condition on each local
  //equation in 'globalEqn'. This means, that the following modification
  //should be made to A and b, for each globalEqn:
  //
  //for(each local equation i){
  //   for(each column j in row i) {
  //      if (i==j) b[i] = gamma/alpha;
  //      else b[j] -= (gamma/alpha)*A[j,i];
  //   }
  //}
  //
  //all of row 'globalEqn' and column 'globalEqn' in A should be zeroed,
  //except for 1.0 on the diagonal.
  //
  double      rhs_term,temp; bool lamb;
  int         ierr,self,jj,ii,kk,*colids,*bids,ncols,nloc,proc;
  int         myeqi, geqi, lidi, ndfi, colj;
  const int   my0peq = proc_primEq_[mype_];
  PromGrid   *grid0 = prom_->grid0_;   assert(grid0!=NULL);
  PromMatrix *KK = grid0->stiffness_; 
  int       **idsarr=NULL,*localrows=NULL,*sizes=NULL;
  PRINT1(len);

  if( verbose_ > 4 || (mype_==0 && verbose_==3) ){
    PromContext::printArr(stderr,globalEqn,len,
			  "enforceEssentialBC::globalEqn[%d]");
    PromContext::printArr(stderr,gamma,len,
			  "enforceEssentialBC::gamma[%d]");
    PromContext::printArr(stderr,alpha,len,
			  "enforceEssentialBC::alpha[%d]");
  }

  // test 
  {  const int my0eq=proc_globalEq_[mype_], next0eq=proc_globalEq_[mype_+1];
  for( ii = 0, nloc = 0 ; ii < len; ii++ ) {
    geqi = globalEqn[ii]; 
    if( geqi >= my0eq && geqi < next0eq ) nloc++;
  }}
  if( nloc == 0 ) return 0;

  // make zero-eqs with these
  ierr = PetscMalloc(nloc*sizeof(int),&localrows); CHKERRQ(ierr);
  ierr = PetscMalloc(nloc*sizeof(int),&sizes); CHKERRQ(ierr);
  ierr = PetscMalloc(nloc*sizeof(int*),&idsarr); CHKERRQ(ierr);
  for( ii = 0, nloc = 0 ; ii < len ; ii++ ) {
    rhs_term = gamma[ii]/alpha[ii]; // alpha is 1.0 for 'normal' BCs
    alpha[ii] = rhs_term; // output!!!
    geqi = globalEqn[ii]; 
    ierr = eq_map_private( geqi, &myeqi, &lamb, &proc ); CHKERRQ(ierr);
    if( !lamb ){ // Alan uses 'lamb'
      if( proc == mype_ ) { // use only local
	//  b[j] -= (gamma/alpha)*A[j,i]; done latter
	if( rhs_term != 0.0 ){
	  temp = rhs_term; kk = myeqi;
	  ierr = bmod->SetValues(1, &kk, &temp, INSERT_VALUES); CHKERRQ(ierr);
	}
	ierr = primGEq_lid_private( myeqi, &lidi ); CHKERRQ(ierr);
	ndfi = (lid_primLEq_[lidi+1] - lid_primLEq_[lidi]); 
	// get column ids in this row
	ierr = KK->GetRow( myeqi, &ncols, &colids ); CHKERRQ(ierr);
	if( ncols <= 0 ) SETERRQ3(1,"cols (%d) <= 0 eq=%d,%d -- Check FEAP files (ie, 'mate' for correctness, if using FEAP",ncols,ii,geqi); 
	ierr = PetscMalloc(ncols*sizeof(int),&bids); CHKERRQ(ierr);
	// make the rhs modification here.
	for( jj = self = kk = 0; jj < ncols; jj++ ) {
	  colj = colids[jj];
	  if( colj == myeqi ) self++;
	  else bids[kk++] = colj;
	} // end for(jj<rowLength) loop
	if( self != 1 ) SETERRQ2(1,"self != 1 (%d) eq=%d",self,myeqi); 
	if( kk != ncols-1 ) SETERRQ(1,"kk != ncols-1"); 
	sizes[nloc] = ncols - 1; idsarr[nloc] = bids;
      	// keep for zero matrix cols and rows
	localrows[nloc] = myeqi; 
	// set fixity in Prometheus nodes
	jj = myeqi - lid_primLEq_[lidi] - my0peq; assert(jj>=0 && jj<ndfi);
	grid0->nodes_[lidi]->setFixed( jj, TRUE ); 
	ierr = KK->PromMatrix::RestoreRow( myeqi, &ncols, &colids ); 
	CHKERRQ(ierr);
	nloc++;
      }
    }
  }

  // zero rows and cols -- cache
  Prometheus_ZeroEquations *zeqs =
    new Prometheus_ZeroEquations( localrows, sizes, idsarr, nloc );
  CHKNEW(zeqs,1);
  zeqs->next_ = zeroEqs_;
  zeroEqs_ = zeqs;
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::enforceRemoteEssBCs"
/* Prometheus_LinSysCore::enforceRemoteEssBCs() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::enforceRemoteEssBCs( int numEqns, 
						int* globalEqns,
						int** colIndices, 
						int* colIndLen,
						double** coefs ) 
{
  //
  //globalEqns should hold eqns that are owned locally, but which contain
  //column indices (the ones in colIndices) which are from remote equations
  //on which essential boundary-conditions need to be enforced.
  //
  //This function will only make the modification if the above conditions
  //hold -- i.e., the equation is a locally-owned equation, and the column
  //index is NOT a locally owned equation.
  //
  PRINT2(numEqns,*globalEqns); 
  
  // ** this is done in enforceEssentialBC **
  /*for( ii=0; ii<numEqns; ii++) {
    tt = globalEqns[ii];
    if( tt >= my0eq && tt < next0eq ) {
    int ncols = colIndLen[ii], *coleqs = colIndices[ii];
    ierr = eq_map_private( tt, &myeqi, &lamb, &proc ); CHKERRQ(ierr);
    // set prom_->grid0_->nodes[myeqi...-my0].setFixed(dof,ndof_)
    if( lamb ){ SETERRQ(1,"Lagrange mult eq"); }
    if( proc != mype_){ SETERRQ(1,"proc!=myproc"); }
    //const int lidi = tt;
    for( jj = 0; jj < ncols; jj++ ) {
    tt = coleqs[jj];
    ierr = eq_map_private( tt, &myeqj, &lamb, &proc ); CHKERRQ(ierr);
    if( proc == mype_ ) continue;
    if( lamb ){ SETERRQ(1,"Lagrange mult eq"); }
    // do not have ghost-eq to ghost map -- ugly search for myeqj!
    PromNode *curr = NULL;
    for(int xx = grid->nLocalNd_ ; xx < grid->nNodes_; xx++ ){
    curr = grid->nodes_[xx]; const int geq = curr->geq(); 
    if( myeqj >= geq && myeqj < geq + curr->ndf() ) break;
    else curr = NULL;
    }
    if( curr == NULL ){ SETERRQ1(1,"curr==NULL",myeqj); }
    //curr->setFixed( myeqj - curr->geq(), TRUE );
    }
    }
    }*/

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::enforceOtherBC_private"
/* Prometheus_LinSysCore::enforceOtherBC_private() ***************************
 *
 *  - Enforces "other" (ie, kindof odd) BCs  
 *
 *  INPUT: 
 *
 *
 *  SIDE EFFECTS:
 *   - Adds to A and b
 */
int Prometheus_LinSysCore::enforceOtherBC_private( int* globalEqn, 
						   double* alpha,
						   double* beta, 
						   double* gamma,
						   int len) 
{
  //
  //This function must enforce a natural or mixed boundary condition on the
  //equations in 'globalEqn'. This means that the following modification 
  //should be made to A and b:
  //
  //A[globalEqn,globalEqn] += alpha/beta;
  //b[globalEqn] += gamma/beta;
  //
  bool lamb; assert(currentRHS_>=0);
  PromVector *BB = bVecArr_[currentRHS_];
  PromVector *b = BB;   
  PromMatrix *KK = prom_->grid0_->stiffness_;
  int ierr,myeqi,geqi,ii,nloc,*localrows=NULL,proc;
  const int my0eq = proc_globalEq_[mype_], next0eq = proc_globalEq_[mype_+1];
  double temp, *bvec = NULL, fbuff[PROM_BSZ]; int rbuf[PROM_BSZ];
  PRINT2(len,*globalEqn); 

  for( ii=0,nloc=0; ii<len; ii++) {
    geqi = globalEqn[ii];
    //if globalEqn[i] is local, we'll diagonalize the row and column.
    if( geqi >= my0eq && geqi < next0eq ) nloc++;
    //else SETERRQ(1,"off proc eq. ???"); 
  }
  if( nloc == 0 ) return 0;
  PRINT2(len,*globalEqn);

  if( nloc > PROM_BSZ ) {
    ierr = PetscMalloc(nloc*sizeof(int),&localrows); CHKERRQ(ierr);
    ierr = PetscMalloc(nloc*sizeof(double),&bvec); CHKERRQ(ierr);
  }
  else { localrows = rbuf; bvec = fbuff; }
  
  for( ii = 0, nloc = 0 ; ii < len; ii++ ) {
    geqi = globalEqn[ii];
    if( geqi >= my0eq && geqi < next0eq ) {
      //if globalEqn[i] is local, we'll diagonalize the row and column.
      ierr = eq_map_private( geqi, &myeqi, &lamb, &proc ); CHKERRQ(ierr);   
      // vec
      localrows[nloc] = myeqi; // keep for rows
      bvec[nloc] = gamma[ii]/beta[ii];   // b_[i] += gamma[i]/beta[i];*/
      // 
      temp = alpha[ii]/beta[ii]; // coefs[j] += alpha[i]/beta[i];
      if( temp != 0.0 ) {
	ierr = KK->PromMatrix::SetValues(1,&myeqi,1,&myeqi,&temp,ADD_VALUES);
	CHKERRQ(ierr);
      }
      nloc++;
    }
  }
  // fix 
  ierr = b->SetValues( nloc, localrows, bvec, ADD_VALUES ); CHKERRQ(ierr);

  if( localrows != rbuf ) { PetscFree(localrows); PetscFree(bvec); }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::collectBCs"
/* Prometheus_LinSysCore::collectBCs() ***************************
 *
 *  - collective, calls MatAssembly
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 *   - set fixity in Prometheus nodes
 *   - Adds to A and b
 *   - adds to 'gamma_alpha_'
 *   - create 'zeroIS_'
 */
int Prometheus_LinSysCore::collectBCs()
{
  int  ierr;
  PromGrid   *grid0 = prom_->grid0_;
  PromMatrix *AA = grid0->stiffness_;  assert(AA!=NULL);
  MPI_Comm wcomm = MPIComm_;

  PromIS zis = (PromIS)zeroIS_;
  if( zis != NULL ){ 
    ierr = PromISDestroy( zis ); CHKERRQ(ierr); 
    zis = NULL; int nnodes = grid0->nNodes_;
    // clear node fixity
    for( int ii = 0 ; ii < nnodes ; ii++ ) {
      PromNode *curr = grid0->nodes_[ii]; int ndf = curr->getNDF();
      for( int i = 0 ; i < ndf ; i++ ) curr->setFixed( i, FALSE ); 
    }
  }
  // OtherBC: calls MatSetValues (diagonal only), ADD_VALUES
  for( Prometheus_OthBCData *obc = othbcdata_, *obc2 ; obc ; obc = obc2 ) {
    // adds values on diag of A and b local
    ierr = enforceOtherBC_private( obc->globalEqn_, obc->alpha_, obc->beta_,
				   obc->gamma_, obc->len_ ); 
    CHKERRQ(ierr);
    obc2 = obc->next_; delete obc;  
  }
  othbcdata_ = NULL;
  if(verbose_>1)PetscPrintf(wcomm,"[%d]%s Matrix Assemble\n",mype_,__FUNCT__);
  // assemble -- collective, uses getrow
  ierr = AA->Assembly();   CHKERRQ(ierr);
  if(verbose_>1)PetscPrintf(wcomm,"[%d]%s Matrix Assemble Done\n",mype_,__FUNCT__);
  // calls getrow, make zeroEqs_, collect non-zero Dir. BC's in "gamma_alpha"
  { 
    ierr = gamma_alpha_->Set( 0.0 ); CHKERRQ(ierr);
    Prometheus_EssBCData *esbc, *esbc2;
    for( esbc = essbcdata_ ; esbc ; esbc = esbc2 ) {
      // uses getrow (needs to be assembled), adds into b global, sets node.fix
      ierr = enforceEssentialBC_private(esbc->globalEqn_, esbc->alpha_,
					esbc->gamma_,esbc->len_, gamma_alpha_);
      CHKERRQ(ierr);
      esbc2 = esbc->next_;      delete esbc;
    }
    essbcdata_ = NULL;
  }

  { // form 'zoroIS_'
    Prometheus_ZeroEquations *zeqs;
    int ii,jj,nrows,*rows,rbuf[PROM_BSZ];
    // count
    for( nrows = 0, zeqs = zeroEqs_ ; zeqs ; zeqs = zeqs->next_ ) {
      int nloc = zeqs->len_;
      nrows += nloc;
    }
    // zero rows -- set diag == 1.0  
    if( nrows > PROM_BSZ){ 
      ierr = PetscMalloc(nrows*sizeof(int),&rows); CHKERRQ(ierr);
    }
    else rows = rbuf;
    for( ii = 0, zeqs = zeroEqs_ ; zeqs ; zeqs = zeqs->next_ ) {
      int nloc = zeqs->len_;
      for( jj = 0; jj < nloc; jj++, ii++ ){
	rows[ii] = zeqs->zeroRowsLeq_[jj];
      }
    }
    assert(ii == nrows);
    qsort( (void*)rows, nrows, sizeof(int), prom_intcompar );
    
    // zero rows with old diagonal -- collective
    ierr = PromISCreateGeneral(PETSC_COMM_SELF,nrows,rows,&zis);CHKERRQ(ierr);
    zeroIS_ = zis;

    if( nrows ) {
      if( rows != rbuf ) PetscFree(rows); 
      rows = NULL;
    }
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::assembleCRs"
/* Prometheus_LinSysCore::assembleCRs() ***************************
 *
 *  - assemble cached CRs into 'uzawaC_' and 'C_' or 'KK->C_' 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::assembleCRs( PromTable<int> *CReq_projCReq ) 
{
  PromMatrix *KK = prom_->grid0_->stiffness_;
  PromMatrix *Cproj = use_uzawa_ ? C_ : NULL;
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); 
  if(K==NULL && uzawaC_==NULL && Cproj==NULL) { 
    assert(CRCache_==NULL); return 0; 
  }
  PromMatrix *CC = use_uzawa_ ? uzawaC_ : K->C_;
  PromMatrix *Ct = use_uzawa_ ? NULL : K->Ct_;
  assert( Cproj != NULL || CC != NULL );         assert(currentRHS_>=0);
  PromVector *B = bVecArr_[currentRHS_];
  PromCRVector *BB = dynamic_cast<PromCRVector*>(B); assert(BB!=NULL);
  PromVector *gg = BB->p_; assert(gg != NULL);
  int ierr,jj; PromCRCache *crc;
  PRINT;

  // " FEI: pt. assem."
  prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FEI_ASSEM]); 
  //  
  for( crc = CRCache_ ; crc != NULL ; /*void*/ ) {
    // apply BCs to constraints here
    if( geq_fixity_ != NULL ) {
      for(int jj=0;jj<crc->len_;jj++) {
	int geq = crc->geqs_[jj]; double val = crc->vals_[jj];
	if( val != 0.0 ) { 
	  double bc = geq_fixity_->findd( geq + 1 ); 
	  if( bc != PROM_DTAB_ZERO && crc->vals_[jj] != 0.0 ) {
	    // move to RHS
	    if( bc != 0.0 ) {
	      double vv = -crc->vals_[jj] * bc;
	      ierr = gg->SetValues(1,&crc->geq_,&vv,ADD_VALUES); CHKERRQ(ierr);
	    }
	    crc->vals_[jj] = 0.0;
	  }
	}
      }
    }
    double *vv=crc->vals_; int *ids=crc->geqs_, nn = crc->len_,eq=crc->geq_;
    ierr = CC->SetValues( 1, &eq, nn, ids, vv, INSERT_VALUES ); CHKERRQ(ierr);
    //
    if( Ct != NULL ) {
      ierr = Ct->SetValues(nn, ids, 1, &eq, vv, INSERT_VALUES); CHKERRQ(ierr);
    }
    // inc
    PromCRCache *crc2 = crc->next_; delete crc; crc = crc2;
  }
  CRCache_ = NULL;
  if( CC != NULL ) {
    ierr = CC->Assembly();   CHKERRQ(ierr); 
    ierr = CC->SetDirty( FALSE ); CHKERRQ(ierr); 
  }  
  if( Ct != NULL ) {
    ierr = Ct->Assembly();   CHKERRQ(ierr); 
    ierr = Ct->SetDirty( FALSE ); CHKERRQ(ierr); 
  }

  // proj
  if( Cproj != NULL ) {    
    assert(use_uzawa_);
    // assemble but do not destroy cache
    int projEq = Cproj->map_.my0Eq(), eq;
    for( crc = projCRCache_; crc != NULL ; /*void*/ ) {
      // apply BCs to constraints here
      if( geq_fixity_ != NULL ) {
	for(jj=0;jj<crc->len_;jj++) {
	  int geq = crc->geqs_[jj]; double val = crc->vals_[jj];
	  if( val != 0.0 ) { 
	    double bc = geq_fixity_->findd( geq + 1 ); 
	    if( bc != PROM_DTAB_ZERO ) {
	      // move to RHS
	      if( bc != 0.0 ) {
		double vv = -val * bc;
		ierr = gg->SetValues(1,&crc->geq_,&vv,ADD_VALUES);
		CHKERRQ(ierr);
	      }
	      crc->vals_[jj] = 0.0;
	    }
	  }
	}
      }
      assert(CReq_projCReq != NULL);
      if( (eq=CReq_projCReq->find(crc->geq_+1)) ) eq--;
      else {
	CReq_projCReq->add( projEq+1, crc->geq_+1 );
	eq = projEq++;
      }
      ierr = Cproj->SetValues( 1, &eq, crc->len_, crc->geqs_, crc->vals_,
			       INSERT_VALUES );
      CHKERRQ(ierr);
      PromCRCache *crc2 = crc->next_; delete crc; crc = crc2;
    }
    projCRCache_ = NULL;   
    ierr = Cproj->Assembly();   CHKERRQ(ierr); 
    ierr = Cproj->SetDirty( FALSE ); CHKERRQ(ierr); 
  }
  prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FEI_ASSEM]);

  if( CC != NULL ) {
    // C <- C D^{-1/2}
    if( prom_->right_ != NULL && prom_->left_ != NULL ) { // broken!
      ierr = CC->Scale( prom_->left_, prom_->right_ );    CHKERRQ(ierr);
      if( Cproj != NULL ) {
	ierr = Cproj->Scale(prom_->left_, prom_->right_); CHKERRQ(ierr);
      }
    }
    else assert(prom_->right_ == NULL && prom_->left_ == NULL);
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::applyBCs"
/* Prometheus_LinSysCore::applyBCs() ***************************
 *
 *  - applies BCs to matrix (A and C) and adjusts RHS for non-zero BCs
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::applyBCs( PromMatrix *projC ) 
{
  double temp;
  int ierr,nrows,jj;
  PromGrid *grid0 = prom_->grid0_;
  PromMatrix *KK = grid0->stiffness_;
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); 
  PromMatrix   *CC = use_uzawa_ ? uzawaC_ : (K==NULL ? NULL : K->C_);
  MPI_Comm wcomm = MPIComm_;
  PRINT;

  // adjust rhs
  ierr = gamma_alpha_->NormMax( &temp ); CHKERRQ(ierr); 
  if( temp != 0.0 ) {
    assert(currentRHS_>=0); 
    PromVector *BB = bVecArr_[currentRHS_]; assert(BB!=NULL); 
    if(verbose_>1)PetscPrintf(wcomm,"[%d]%s Have non-zero BCs\n",mype_,__FUNCT__);
    ierr = KK->PromMatrix::Mult(gamma_alpha_, work_);  CHKERRQ(ierr);
    ierr = BB->PromVector::AXPY(-1.0, work_); CHKERRQ(ierr); // b <-- b - A*bc
    if( CC != NULL || projC != NULL ) {
      PromCRVector *B = dynamic_cast<PromCRVector*>(BB); assert(B!=NULL);
      PromCRVector *W = dynamic_cast<PromCRVector*>(work_); assert(W!=NULL);
      assert(W->p_ != NULL && B->p_ != NULL);
      if( CC != NULL ) ierr = CC->Mult( gamma_alpha_, W->p_ ); 
      else ierr = projC->Mult( gamma_alpha_, W->p_ );  CHKERRQ(ierr);
      ierr = B->p_->AXPY( -1.0, W->p_ );   CHKERRQ(ierr);
    }
  }

  // zero rows
  if( zeroIS_ == NULL ){ SETERRQ(1,"zeroIS_ == NULL"); }
  PromIS zis = (PromIS)zeroIS_; 
  //KK->PromMatrix::NormFrob( &temp ); CHKERRQ(ierr);
  ierr = KK->GetDiagonal( work_ );      CHKERRQ(ierr);
  ierr = work_->PromVector::NormMax( &temp );        CHKERRQ(ierr);
  ierr = KK->ZeroRows( zis, temp*PROM_DIAG_FACTOR ); CHKERRQ(ierr);
  // zero cols, no diagonal, INSERT_VALUES
  ierr = PromISGetSize( zis, &nrows ); CHKERRQ(ierr);
  if( nrows > 0 ) {
    if( zeroEqs_ == NULL ){ SETERRQ(1,"zeroEqs_ == NULL"); }
    double fbuff__[PROM_BSZ], *zeros = fbuff__;
    int neqs = zeroEqs_->len_, zsize = 0;
    for( jj = 0; jj < neqs; jj++ ) {
      int sz = zeroEqs_->sizes_[jj];   
      if( sz > zsize ){
	if( sz > PROM_BSZ ) {
	  if( zeros != fbuff__ ) PetscFree( zeros ); 
	  ierr = PetscMalloc( sz * sizeof(double), &zeros ); CHKERRQ(ierr);
	  for( int ii = 0; ii < sz; ii++ ) zeros[ii] = 0.0;
	  zsize = sz;
	}
	else {
	  do zeros[zsize++] = 0.0; while(zsize < sz);
	}
      }
      int col = zeroEqs_->zeroRowsLeq_[jj]; // 'col' is local!?!?!
      int *ids = zeroEqs_->idsarr_[jj];
      ierr = KK->SetValues(sz, ids, 1, &col, zeros, INSERT_VALUES);
      CHKERRQ(ierr);
      if( KK->isComplex() ) {
	PromComplexMatrix *K = dynamic_cast<PromComplexMatrix*>(KK); 
	assert(K!=NULL);
	ierr = K->SetValues_i(sz, ids, 1, &col, zeros, INSERT_VALUES);
	CHKERRQ(ierr);
      }
      // PetscFree(ids); // deleting zeros
    }
    assert( zeroEqs_->next_ == NULL ); 
    if( zeros != fbuff__ ) PetscFree( zeros ); 
  }

  // assemble -- collective
  ierr = KK->FlushAssembly();   CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::getRHS"
/* Prometheus_LinSysCore::getRHS() ***************************
 *
 *  - get pointer to RHS
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
PromVector * Prometheus_LinSysCore::getRHS()
{
  return bVecArr_[currentRHS_];
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::getSolution"
/* Prometheus_LinSysCore::getSolution() ***************************
 *
 *  - get pointer to Solution
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
PromVector *Prometheus_LinSysCore::getSolution()
{
  return xVec_;
}

/************************************************************************
 * Solve
 ************************************************************************/

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::launchSolver"
/* Prometheus_LinSysCore::launchSolver() ***************************
 *
 *  - Solve
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 *   - changes RHS if BCs conflict
 *
 */
int Prometheus_LinSysCore::launchSolver( int& solveStatus, int& iterations ) 
{
  PRINT;
  // This function does any last-second setup required for the
  // linear solver, then goes ahead and launches the solver to get
  // the solution vector.
  // Also, if possible, the number of iterations that were performed
  // is stored in the iterations_ variable.
  //
  // AZ_normal      0   (successful solve)
  // AZ_param      -1   (means a requested option isn't available)
  // AZ_breakdown  -2   (numerical breakdown during iteration)
  // AZ_maxits     -3   (means iteration limit was reached)
  // AZ_loss       -4   (loss of precision)
  // AZ_ill_cond   -5   (GMRES Hessenberg is ill-conditioned)
  PromMatrix *KK = prom_->grid0_->stiffness_;
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); 
  PromMatrix   *CC = use_uzawa_ ? uzawaC_ : (K==NULL ? NULL : K->C_);
  PromVector *XX = xVec_; PromCRVector *X = dynamic_cast<PromCRVector*>(XX);
  PromCRVector *W = dynamic_cast<PromCRVector*>(work_);
  PromVector *pp = (X==NULL) ? NULL : X->p_; 
  const PromVector * const BB = bVecArr_[currentRHS_];
  const PromCRVector * const B = dynamic_cast<const PromCRVector*>(BB);
  MPI_Comm wcomm = MPIComm_;
  int ierr,its,ret,fact = (KK->isComplex() ? 2 : 1); 
  PromSolver *const solver = prom_->solver_;      assert(solver!=NULL);
  float normR; double dnorm;

  if( KK->isDirty() ) { 
    solveStatus = -1; SETERRQ(1,"A->isDirty()"); 
  }
  if( CC && CC->isDirty() ) { 
    solveStatus = -1; SETERRQ(1,"C->isDirty()"); 
  }

  // output matrices and vects
  nSolves_++;
  if( verbose_ > 2 ) {
    double *vv;    FILE *file;
    int ii,jj;
    const int my0prim = proc_primEq_[mype_], next0prim=proc_primEq_[mype_+1];
    const int nlocPrims = (next0prim - my0prim)*fact; // first local Lm eq
    char path[128]; 
    PetscPrintf(wcomm,"\t[%d] %d) output matrices and vectors path=%s\n",
		mype_, nSolves_, root_ );
    // A
    ierr = PromContext::WriteMat( KK, "A_", nSolves_ ); CHKERRQ(ierr);
    // b
    if( npe_ == 1 ) sprintf( path, "%s/b_%d", root_, nSolves_ );
    else sprintf( path, "%s/b_%d_%d", root_, nSolves_, mype_);
    file = fopen( path, "w");
    ierr = BB->GetArray( &vv );    CHKERRQ(ierr);
    if( BB->isComplex() ) {
      for( ii=0, jj=my0prim ; ii < nlocPrims ; ii += 2, jj++){
	if(vv[ii] != 0.0 || vv[ii+1] != 0.0 ) 
	  fprintf(file,"%d %le %le\n",jj+1,vv[ii],vv[ii+1]); 
      }
    }
    else {
      for(ii=0,jj=my0prim;ii<nlocPrims;ii++,jj++){
	if(vv[ii] != 0.0) fprintf(file,"%d %le\n",jj+1,vv[ii]); 
      }
    }
    ierr = BB->RestoreArray( &vv ); CHKERRQ(ierr); 
    fclose(file);
    // C
    if( CC != NULL ) {
      const PromVector *gg = B->p_;
      const int my0Lm = proc_lmEq_[mype_];
      const int nlocLm = proc_lmEq_[mype_+1] - my0Lm;
      if( npe_ == 1 ) sprintf( path, "%s/C_%d", root_, nSolves_);
      else sprintf( path, "%s/C_%d_%d", root_, nSolves_, mype_);
      file = fopen( path, "w");
      ierr = PromContext::WriteMat_private( CC, file, KK->getM(), 0, FALSE );
      CHKERRQ(ierr);
      fclose(file);
      // C'
      if( npe_ == 1 ) sprintf( path, "%s/Ct_%d", root_,nSolves_);
      else sprintf( path, "%s/Ct_%d_%d", root_, nSolves_, mype_ );
      file = fopen( path,  "w" );
      ierr = PromContext::WriteMat_private( CC, file, 0,  KK->getM(), TRUE );
      CHKERRQ(ierr);
      fclose(file);
      // g
      if( npe_ == 1 ) sprintf(path, "%s/g_%d", root_, nSolves_);
      else sprintf( path, "%s/g_%d.%d", root_, nSolves_, mype_);
      file = fopen( path, "w");
      ierr = gg->GetArray( &vv );    CHKERRQ(ierr);
      for( ii = 0, jj = my0Lm + KK->getM() ; ii < nlocLm ; ii++, jj++ ){
	if(vv[ii] != 0.0) fprintf(file,"%d %le\n",jj+1,vv[ii]); 
      }
      ierr = gg->RestoreArray( &vv ); CHKERRQ(ierr); fclose(file);
    }
    if( solver->projector_ != NULL ) {
      const PromMatrix *proj = solver->projector_->projC_;
      if( npe_ == 1 ) sprintf( path, "%s/projC_%d", root_, nSolves_);
      else sprintf( path, "%s/projC_%d_%d", root_, nSolves_, mype_);
      file = fopen( path, "w");
      ierr = PromContext::WriteMat_private(proj,file,0,0,FALSE); CHKERRQ(ierr);
      fclose(file);
    }
  }
  //
  prom_->perf_mon_.StagePush(PromContext::prom_stages_[0]);
  prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FEI_SOLVE]);
  ret = solveStatus = iterations = 0;  

  // copy 'work_' <-- B (and scale)
  ierr = work_->Copy( BB ); CHKERRQ(ierr);
  // scale work
  if( prom_->left_ != NULL ) {
    // scale f <- f * D^{-1/2}
    ierr = work_->PromVector::PointwiseMult(work_, prom_->left_);
    CHKERRQ(ierr);
  }

  // get 'dirty_guess_'
  its = dirty_guess_ ? 1 : 0;
  MPI_Allreduce( &its, &ierr, 1, MPI_INT, MPI_MAX, wcomm );
  dirty_guess_ = (ierr == 0) ? FALSE : TRUE;  

  { // put init. guess can conflict in fei test 'share3BC'
    double *bc,*ig,*bb; int ii, geq, nleq = XX->getn()*fact;
    if( zeroIS_ == NULL ){ SETERRQ(1,"zeroIS_ == NULL"); } 
    PromIS zis = (PromIS)zeroIS_;  int *bcids, nbc, id; 
    ierr = PromISGetSize( zis, &nbc ); CHKERRQ(ierr); 
    ierr = PromISGetIndices( zis, &bcids ); CHKERRQ(ierr); 
    ierr = gamma_alpha_->GetArray( &bc );    CHKERRQ(ierr);
    ierr = init_guess_->GetArray( &ig );    CHKERRQ(ierr);
    ierr = work_->GetArray( &bb );    CHKERRQ(ierr);
    for(id=ii=0, geq = proc_primEq_[mype_] ; ii < nleq ; ii += fact, geq++){
      if( id < nbc && bcids[id] == geq ) {
	id++;
	if( ig[ii] != 0.0 ) { 
	  if( verbose_ > 1 ) {
	    PetscPrintf(PETSC_COMM_SELF,"\tPrometheus Warning: init guess and BC conflict B[%d]=%e ig=%e\n",geq,bb[ii],ig[ii]);
	  }
	  ig[ii] = 0.0; 
	}
	if( bb[ii] != 0.0 ) {
	  // conflict w/ BC: p1ut BC in true RHS, for residual, and zero working RHS
	  PromVector *B = bVecArr_[currentRHS_]; double z = -bb[ii];
	  ierr = B->SetValues(1, &geq, &z, ADD_VALUES ); CHKERRQ(ierr);
	  if( verbose_ > 1 ) {
	    PetscPrintf(PETSC_COMM_SELF,
			"\tPrometheus Warning: RHS and BC conflict B[%d]=%e bc=%e\n",
			geq,bb[ii],bc[ii]);
	  }
	  bb[ii] = 0.0;
	}
      }
    }
    ierr = PromISRestoreIndices( zis, &bcids ); CHKERRQ(ierr); 
    ierr = gamma_alpha_->RestoreArray( &bc );    CHKERRQ(ierr);
    ierr = init_guess_->RestoreArray( &ig );    CHKERRQ(ierr);
    ierr = work_->RestoreArray( &bb );    CHKERRQ(ierr);
  }

  // get 'normR', apply init guess, change rtol
  bool use_p_init_guess = FALSE; //(pp != NULL);
  const double rtol_orig = solver->getRTol(); 
  double rtol = rtol_orig;
  if( dirty_guess_ || use_p_init_guess ) {
    if(solver->projector_!=NULL) 
      PetscPrintf(wcomm,"Prometheus Warning: initial guess w/ projection CRs not fully supported\n");
    // get 'normR'
    double nR;
    if( rhs_res_scale_ == 1 || use_p_init_guess ) {
      ierr = W->Norm2( &nR );  CHKERRQ(ierr);
      assert(0); // not supported
    }
    // scale
    if( prom_->right_ != NULL ) {
      // scale x <- x * D^{1/2}
      ierr = init_guess_->PointwiseDiv( init_guess_, prom_->right_ );
      CHKERRQ(ierr);
    }
    if( dirty_guess_ ) {
      // r <- b - A x'; (g <- g - C x' (done in assemble CRs)
      ierr = KK->Mult( init_guess_, XX ); CHKERRQ(ierr);
      ierr = work_->AXPY( -1.0, XX ); CHKERRQ(ierr); 
    }
    // p init guess from last solution!
    if( use_p_init_guess ) {
      ierr = CC->MultTranspose( pp, XX );   CHKERRQ(ierr);
      ierr = work_->PromVector::AXPY( -1.0, XX ); CHKERRQ(ierr); 
      //ierr = init_guess_->p_->Copy(pp);CHKERRQ(ierr);//save init. guess for p
      assert(0); // not supported
    }
    //
    ierr = work_->Norm2( &dnorm );   CHKERRQ(ierr); 
    // use original (RHS) to scale residual, by scaling test
    if( rhs_res_scale_ == 1 || use_p_init_guess ) {
      rtol = rtol_orig * nR / dnorm; // change tol.
      solver->SetRTol( rtol ); 
      if( nR < dnorm ){
	PetscPrintf(wcomm,"Prometheus Warning: |b|=%e < |b - Ax|=%e\n",
		    nR, dnorm );
      }
      normR = nR;
    }
    else normR = dnorm;
  }
  else {
    // get 'normR'
    ierr = work_->Norm2( &dnorm ); CHKERRQ(ierr); 
    normR = dnorm;
  }

  // check for zero residual
  if( normR <= normRTol_ ) {
    if(verbose_>0)PetscPrintf(wcomm,"\t[%d]%s Converged %e <= %e\n",
			      mype_,__FUNCT__,normR,normRTol_);
    if( dirty_guess_ ) { // add init guess in
      ierr = XX->PromVector::AXPY( 1.0, init_guess_ ); CHKERRQ(ierr);
    }
    solveStatus = ret;
    dirty_guess_ = FALSE;
    //
    prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FEI_SOLVE]);
    prom_->perf_mon_.StagePop();
    return 0;
  }

  // solve A x = 'work_
  iterations = 0; 
  if( CC == NULL || !use_uzawa_ ) { 
    // solve
    ierr = XX->Set(0.0);   CHKERRQ(ierr);
    ierr = solver->Solve( work_, XX, &its, TRUE, FALSE );
    if( ierr < 0 ) CHKERRQ(ierr); 
    if( its < 0 ) { ret = -3; its = -its; }
    else ret = 0;
    if( dirty_guess_ ) { // add init guess in
      ierr = XX->PromVector::AXPY( 1.0, init_guess_ ); CHKERRQ(ierr);
    }
    if( use_p_init_guess ) {
      //ierr = XX->p_->AXPY( 1.0, init_guess_->p_ ); CHKERRQ(ierr);
    }
  }
  else { // Uzawa
    // solve
    ierr = SolveUzawa(solver,KK,W,X,B,init_guess_,verbose_,normR,ret,its);
    CHKERRQ(ierr);
  }

  solver->SetRTol( rtol_orig );
  iterations = its;
#if defined(PETSC_USE_BOPT_g)
  if( !dirty_guess_ ) { 
    if( zeroIS_ == NULL ){ SETERRQ(1,"zeroIS_ == NULL"); } 
    PromIS zis = (PromIS)zeroIS_; double *xx; 
    int*bcids,nbc,my0eq=proc_primEq_[mype_];
    ierr = PromISGetSize( zis, &nbc ); CHKERRQ(ierr); 
    ierr = PromISGetIndices( zis, &bcids ); CHKERRQ(ierr);
    ierr = XX->GetArray( &xx );    CHKERRQ(ierr);
    for( int ii = 0 ; ii < nbc ; ii++ ) {
      int leq = fact*(bcids[ii] - my0eq); assert(leq>=0);
      if( xx[leq] != 0.0 ) {
	PetscPrintf(PETSC_COMM_SELF,"Prometheus EERROR: BC eq: x[%d]=%e\n",
		    leq,xx[leq]);
      }
    }
    ierr = XX->RestoreArray( &xx );    CHKERRQ(ierr);
    double beta;
    XX->PromVector::Dot( gamma_alpha_, &beta ); 
    if( beta != 0.0 ) {
      PromContext::printArr(stderr,bcids,nbc,"BC ids[%d]");
      PetscPrintf(wcomm,"Prometheus gamma_alpha' * X = %e\n",beta);
    }
    ierr = PromISRestoreIndices( zis, &bcids ); CHKERRQ(ierr); 
  }
#endif
  // put in non-zero BCs
  ierr = XX->AXPY( 1.0, gamma_alpha_ );    CHKERRQ(ierr);
  ierr = gamma_alpha_->Set( 0.0 );         CHKERRQ(ierr);  

  ierr = init_guess_->Set( 0.0 );  CHKERRQ(ierr);  
  // clean up scaling
  if( prom_->right_ != NULL ) {
    // scale u <- D^{-1/2} * u
    ierr = XX->PromVector::PointwiseMult( XX, prom_->right_ );  
    CHKERRQ(ierr);
  }

  if( verbose_ > 0 && solver->computeSVD() && (CC==NULL || !use_uzawa_) ){
    double emax, emin;
    ierr = solver->ComputeExtremeSVD( its, emax, emin ); CHKERRQ(ierr);
    PetscPrintf( wcomm,
		 "Eigenvalues=(%g,%g) condition of precond. system=%g\n",
		 emin, emax, emax/emin );
    if( emin < -1. ) {
      PetscPrintf( wcomm,"\t\tSwitching to GMRES\n",emin, emax, emax/emin );
      ierr = solver->SetKSPType( PROMKSPGMRES );    CHKERRQ(ierr);
    }
  }

  // explicit residual
  if( verbose_ > 0 ) {
    ierr = LSC_Residual( KK, BB, XX, work_ );    CHKERRQ(ierr);
    ierr = work_->Norm2( &dnorm );               CHKERRQ(ierr);
    if( (dnorm/normR > 1.1*rtol_orig && ret == 0) || 
#if defined(PETSC_USE_BOPT_g)
	true
#else
	false
#endif
	) { 
      if( dnorm/normR > 1.1*rtol_orig && ret == 0 ) ret = -4;

      double lam_norm = 0.0;
      if(CC != NULL){ ierr = W->p_->Norm2( &lam_norm );  CHKERRQ(ierr); }
      PetscPrintf(wcomm, "\t[%d]Prometheus: %d iterations. %s%7.1e (rtol=%7.1e), |r|=%9.3e, |g-Cx|=%7.1e, |b|=%9.3e, %s, %s, residual %s, %s\n", mype_, iterations, (dnorm/normR > 1.1*rtol_orig) ? "Warning: |r|/|b|=" : " |r|/|b|=", dnorm/normR, solver->getRTol(), dnorm, lam_norm, normR, dirty_guess_ ? "have x_0" : "no x_0", (prom_->right_==NULL) ? "no scaling" : "scaling", rhs_res_scale_ ? "scaling RHS" : "scaling R0", (CC == NULL) ? "simple solve" : (use_uzawa_ ? "Uzawa solve" : "KKT all-at-once") );
    }
  }
  else if( verbose_ > 0 ) {
    PetscPrintf( wcomm,"\t[%d]Prometheus: %d iterations.\n", mype_, 
		 iterations);
  }
  solveStatus = ret;
  
  /* auto increase of restart -- Fgmres only
  if( solver->getKSPType()==PROMKSPFGMRES && iterations>solver->nEig() ) {
    int maxit = solver->maxits_, neig;
    if( (neig=solver->nEig()) < maxit ) { // helmholtz heuristic
      neig += 50; if( neig > maxit ) neig = maxit;
      if( neig > solver->nEig() ) {
	ierr = solver->SetNEig( neig ); CHKERRQ(ierr);
	ierr = solver->SetKSPType( PROMKSPFGMRES );    CHKERRQ(ierr);
	if(verbose_>0)PetscPrintf(wcomm,"\t[%d] Increase restart to %d\n",
				  mype_,neig);
      }
    }
    }*/
  if( solver->getKSPType()==PROMKSPFGMRES && iterations>solver->nEig() ) {
    int maxit = solver->maxits_, neig;
    if( (neig=solver->nEig()) < maxit ) { // helmholtz heuristic
      // increase hfact

    }
  }

  // clean up
  dirty_guess_ = FALSE;
  //
  prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FEI_SOLVE]);
  prom_->perf_mon_.StagePop();

  // output
  if( (verbose_ > 0 && nSolves_ == 1) || verbose_ > 1 ) {
    ierr = solver->PrintInfo( npe_ ); CHKERRQ(ierr);
  }
  else if( verbose_ > 0 ) PetscPrintf(wcomm,"   ****\n");

  // output vects
  if( verbose_ > 2 ) { 
    double *vv;    FILE *file;
    int ii,jj;
    const int my0prim = proc_primEq_[mype_], next0prim=proc_primEq_[mype_+1];
    const int nlocPrims = (next0prim - my0prim)*fact; 
    const int my0Lm = proc_lmEq_[mype_];
    const int nlocLm = proc_lmEq_[mype_+1] - my0Lm;
    char path[128]; 
    // x
    if( npe_ == 1 ) sprintf( path, "%s/x_%d", root_, nSolves_ );
    else sprintf( path, "%s/x_%d_%d", root_, nSolves_, mype_);
    file = fopen( path, "w");
    ierr = XX->GetArray( &vv );    CHKERRQ(ierr);
    if( XX->isComplex() ) {
      for(ii=0,jj=my0prim;ii<nlocPrims;ii += 2,jj++){
	if(vv[ii] != 0.0 || vv[ii+1] != 0.0 ) 
	  fprintf(file,"%d %le %le\n",jj+1,vv[ii],vv[ii+1]); 
      }
    }
    else {
      for(ii=0,jj=my0prim;ii<nlocPrims;ii++,jj++){
	if(vv[ii] != 0.0) fprintf(file,"%d %le\n",jj+1,vv[ii]); 
      }
    }
    ierr = XX->RestoreArray( &vv ); CHKERRQ(ierr); fclose(file);

    // p
    if( pp != NULL ) {
      if( npe_ == 1 ) sprintf(path, "%s/p_%d", root_, nSolves_);
      else sprintf( path, "%s/p_%d_%d", root_, nSolves_, mype_);
      file = fopen( path, "w");
      ierr = pp->GetArray( &vv );    CHKERRQ(ierr);
      for( ii = 0, jj = my0Lm + XX->getN() ; ii < nlocLm ; ii++, jj++ ){
	if(vv[ii] != 0.0) fprintf(file,"%d %le\n",jj+1,vv[ii]); 
      }
      ierr = pp->RestoreArray( &vv ); CHKERRQ(ierr); fclose(file);
    }
  }

  /*if( 1 ){
    FILE *file;double *vv;int ii,jj;char path[128]; 
    if( npe_ == 1 ) sprintf( path, "%s/X_%d", root_,nSolves_);
    else sprintf( path, "%s/X_%d_%d", root_, nSolves_, mype_ );
    file = fopen( path,  "w" );
    ierr = work_->GetArray( &vv );    CHKERRQ(ierr);
    for(ii=0,jj=0 ; ii<work_->getn() ; ii++,jj++){
    if(vv[ii] != 0.0) fprintf(file,"%d %le\n",jj+1,vv[ii]); 
    }
    ierr = work_->RestoreArray( &vv ); CHKERRQ(ierr); fclose(file);
    fclose(file);
    if( npe_ == 1 ) sprintf( path, "%s/pX_%d", root_,nSolves_);
    else sprintf( path, "%s/pX_%d_%d", root_, nSolves_, mype_ );
    file = fopen( path,  "w" );
    ierr = XX->GetArray( &vv );    CHKERRQ(ierr);
    for(ii=0,jj=0 ; ii<work_->getn() ; ii++,jj++){
    if(vv[ii] != 0.0) fprintf(file,"%d %le\n",jj+1,vv[ii]); 
    }
    ierr = XX->RestoreArray( &vv ); CHKERRQ(ierr); fclose(file);
    fclose(file);
    }*/

  return 0;
} 

#undef __FUNCT__    
#define __FUNCT__ "PromUzawaSolver::~PromUzawaSolver"
/* PromUzawaSolver::~PromUzawaSolver ***************************
 *
 *  - 
 */
PromUzawaSolver::~PromUzawaSolver()
{ 
  if( uwork_ != NULL ) delete uwork_;
  if( pwork_ != NULL ) delete pwork_;
  if( diag_ != NULL ) delete diag_;
  if( Pen_stiff_ != NULL ) delete Pen_stiff_;
}

#undef __FUNCT__    
#define __FUNCT__ "PromUzawaSolver::SolveUzawa"
/* PromUzawaSolver::SolveUzawa ***************************
 *
 *  - Uzawa solver
 *
 *  INPUT: 
 *   - solver:
 *   - AA:
 *   - RR:
 *   - XX:
 *   - BB: current residual (alias for 'work_')
 *   - x_0: inital guess and scratch 
 *   - verb:
 *   - normR0:
 *   - ret_out:
 *   - its_out:
 *
 *  SIDE EFFECTS:
 */
int PromUzawaSolver::SolveUzawa( PromSolver *const solver,
				 const PromMatrix *const AA,
				 PromCRVector *const RR, 
				 PromCRVector *const XX, 
				 const PromCRVector *const BB, 
				 PromVector *const x_0,
				 const int verb, float normR0,
				 int &ret_out, int &its_out ) const
{
  int ierr, outerits, its; assert(uwork_!=NULL); assert(BB!=RR && RR!=XX);
  PromVector *pp = XX->p_, *uu = XX;
  const PromVector *gg = BB->p_, *ff = BB;
  const PromComm_base &comm = AA->map_.comm_; MPI_Comm wcomm = comm.MPIComm();
  const double rtol=solver->getRTol(),cr_tol=(cr_tol_>0.0)?cr_tol_:rtol;
  double normC, normR, norm; int pind = PromContext::prom_logs_[LOG_AUX1];
  const PromMatrix *const CC = uzawaC_;
  PromVector *Cu_g = pwork_; PromProjKKTSolver *proj = solver->projector_;
  ret_out = its_out = 0;
  
  // add in init guess, then use as temp (p init. guess in already)
  ierr = XX->Copy( x_0 ); CHKERRQ(ierr); 
  solver->SetRTol( (CC->getM()==0) ? rtol : 1.e-2 ); // varibale tol
  for( outerits = 0; outerits < max_uzawa_its_ ; outerits++ ) {
    // projection presolve -- use 'x_0'
    if( proj !=NULL ){
      solver->perf_mon_.EventBegin(pind);
      ierr = proj->PreSolve( AA, RR->p_, RR, x_0 );CHKERRQ(ierr);
      solver->perf_mon_.EventEnd(pind);
    }
    // x <- A^-1 ( f - (A * x + C' * lambda) )
    //ierr = uwork_->Set( 0.0 );                             CHKERRQ(ierr);
    ierr = solver->Solve(RR,uwork_, &its,TRUE,FALSE); if(ierr<0)CHKERRQ(ierr);
    if( its < 0 ) its = -its;
    its_out += its;
    //  
    if( proj != NULL ){
      solver->perf_mon_.EventBegin(pind);
      ierr = proj->PostSolve( x_0, uwork_ ); CHKERRQ(ierr);
      solver->perf_mon_.EventEnd(pind);
    }
    ierr = uu->PromVector::AYPX( 1., uwork_ ); CHKERRQ(ierr); // x_i+1 := x_i +
    // increment lagrange mult. solution -- lam <- lam + Pen*(C*u - g)
    ierr = CC->Mult( uu, Cu_g );      CHKERRQ(ierr);
    ierr = Cu_g->AXPY( -1.0, gg );    CHKERRQ(ierr); // C*u - g
    ierr = Cu_g->Norm2( &normC );     CHKERRQ(ierr);
    ierr = Cu_g->PromVector::PointwiseMult( Cu_g, Pen_stiff_ ); CHKERRQ(ierr);
    ierr = pp->AXPY( 1.0, Cu_g );     CHKERRQ(ierr); // p += Cu - g
    // full residual
    ierr = CC->MultTranspose( pp, uwork_ );   CHKERRQ(ierr);
    if( proj != NULL ) { 
      ierr = proj->ComputeMultiplier( AA, CC, uu, pp, ff ); CHKERRQ(ierr);
      ierr = proj->Apply_Ct_add( pp, uwork_, uwork_ ); CHKERRQ(ierr);
    }
    ierr = AA->MultAdd(uu,uwork_,RR); CHKERRQ(ierr); // (Au + C'p)
    ierr = RR->PromVector::AYPX(-1.0, ff); CHKERRQ(ierr); //r := f - (Au + C'p)

    // test convergence
    ierr = RR->PromVector::Norm2( &normR );  CHKERRQ(ierr); // |f|_2
    norm = sqrt( normR * normR + normC * normC );    // |r|

    if( verb > 0 )
      PetscPrintf(wcomm,
		  "\t[%d] %d) |B-K*X|=%7.1e, %d its (rtol=%7.1e), |g-C*u|=%7.1e (%7.1e <? %7.1e)\n",
		  comm.mype(),outerits+1,norm,its,solver->getRTol(),
		  normC, norm, rtol * normR0 );
    // test convergence 
    if((norm < rtol*normR0 && normC<cr_tol) || outerits+1 == max_uzawa_its_){
      // report last residual
      ierr = solver->Monitor(0,norm); CHKERRQ(ierr);
      break; 
    }
  }

  if( verb > 2 ) {
    PromMatrix *A = (PromMatrix*)AA; // hack to get rid of'const'
    // true residual
    ierr = setRegularization_private(A,CC,diag_,-penalty_factor_);CHKERRQ(ierr);
    ierr = A->Assembly();        CHKERRQ(ierr);
    // primal residual, w/o regularization
    ierr = CC->MultTranspose( pp, uwork_ );   CHKERRQ(ierr);
    if( proj != NULL ) { 
      ierr = proj->ComputeMultiplier( AA, CC, uu, pp, ff ); CHKERRQ(ierr);
      ierr = proj->Apply_Ct_add( pp, uwork_, uwork_ ); CHKERRQ(ierr);
    }
    ierr = AA->MultAdd( uu, uwork_, RR ); CHKERRQ(ierr); // (Au + C'p)
    ierr = RR->PromVector::AYPX( -1.0, ff ); CHKERRQ(ierr); // r := f-(Au+C'p)
    ierr = RR->PromVector::Norm2( &normR );  CHKERRQ(ierr); // |f|_2
    ierr = CC->Mult( uu, Cu_g );      CHKERRQ(ierr);
    ierr = Cu_g->AXPY( -1.0, gg );    CHKERRQ(ierr); // C*u - g
    ierr = Cu_g->Norm2( &normC );     CHKERRQ(ierr);
    norm = sqrt( normR * normR + normC * normC ); 
    ierr = gg->Norm2( &normC );     CHKERRQ(ierr);
    normR = sqrt( normR0 * normR0 + normC * normC ); // full |b| 
    PetscPrintf(wcomm,"\t[%d]Prometheus Uzawa:true res.=%e r-res=%e(tol=%e)\n",
		comm.mype(), norm, norm/normR, rtol );
    ierr = AA->GetDiagonal( diag_ );      CHKERRQ(ierr);
    ierr = setRegularization_private(A,CC,diag_,penalty_factor_);CHKERRQ(ierr);
    ierr = A->Assembly();        CHKERRQ(ierr);
    ierr = A->SetDirty( FALSE ); CHKERRQ(ierr);
  }

  if( outerits == max_uzawa_its_ ) {
    if( verb > 2 ) {
      PetscPrintf(wcomm,
		  "\t[%d]Prometheus WARNING Max Uzawa iterations(%d)\n",
		  comm.mype(), max_uzawa_its_ );
    }
    ret_out = -3; 
  }

  return 0;
} 

#undef __FUNCT__    
#define __FUNCT__ "PromUzawaSolver::Init"
/* PromUzawaSolver::Init ***************************
 *
 */
int PromUzawaSolver::Init( const PromMap &primmap, const PromMap &dualmap )
  {
  int ierr;

  PetscFunctionBegin;
  // pwork
  assert(pwork_ == NULL);
  pwork_ = new PromVector( dualmap ); CHKNEW(pwork_,1);
  ierr = pwork_->Set(0.0); CHKERRQ(ierr);
  // uwork (complex?)
  assert(uwork_ == NULL);
  uwork_ = new PromVector( primmap ); CHKNEW(uwork_,1);
  ierr = uwork_->Set( 0.0 ); CHKERRQ(ierr);
  // diag cache (complex?)
  assert(diag_ == NULL);
  diag_ = new PromVector( primmap );
  // penalty stiffness
  assert(Pen_stiff_ == NULL);
  Pen_stiff_ = new PromVector( dualmap );  CHKNEW(Pen_stiff_,1);
  ierr = Pen_stiff_->Set( PEN_INIT ); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "PromUzawaSolver::Reset"
/* PromUzawaSolver::Reset ***************************
 *
 */
int PromUzawaSolver::Reset( const PromMap &dualmap )
{
  int ierr;
  PetscFunctionBegin;
  // pwork
  assert( pwork_ != NULL );  delete pwork_;
  pwork_ = new PromVector( dualmap ); pwork_->Set( 0.0 );  
  // rest (prim)
  assert( Pen_stiff_ != NULL && uwork_ != NULL ); delete Pen_stiff_;
  Pen_stiff_= new PromVector( dualmap );  Pen_stiff_->Set( PEN_INIT );
  ierr = uwork_->Set( 0.0 ); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::primGEq_lid_private"
/* Prometheus_LinSysCore::primGEq_lid_private() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::primGEq_lid_private(const int geq, int *lid_out)
  const
{
  assert(proc_primEq_!=NULL && lid_primLEq_!=NULL);
  if( geq >= proc_primEq_[mype_] && geq < proc_primEq_[mype_+1] ) {
    const int my0eq = proc_primEq_[mype_], leq = geq - my0eq;
    const int nn = proc_gnode_[mype_+1] - proc_gnode_[mype_];
    const int neq = proc_primEq_[mype_+1] - proc_primEq_[mype_];
    int lid = (int)( (float)nn *(float)leq/(float)neq );
    if( lid < 0 ) lid = 0;
    else if (lid >= nn) lid = nn-1;
    while(1) {
      if( leq < lid_primLEq_[lid] ) lid--; 
      else if( leq >= lid_primLEq_[lid+1] ) lid++;
      else break;
    }
    *lid_out = lid;
  }
  else {
    SETERRQ(1,"Prometheus: Eq must be local!");
  }
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::id_proc_private"
/* Prometheus_LinSysCore::id_proc_private() ***************************
 *
 *  - get proc for global id
 *
 *  INPUT: 
 *   - gid
 *   - proc: (out)
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::id_proc_private( const int gid, int &proc ) const 
{ 
  int nglobal = proc_gnode_[npe_]; 
  if( gid < 0 || gid >= nglobal ){
    PetscPrintf( PETSC_COMM_SELF,
		 "%s line %d in %s ERROR: gid<0||gid>=gnode_proc[numprocs_]\n",
		 __FUNCT__,__LINE__,__FILE__);
    abort();
  }
  if( proc_gnode_[0] != 0 ){
    PetscPrintf( PETSC_COMM_SELF,
		 "%s line %d in %s ERROR: gnode_proc[0] != 0\n",
		 __FUNCT__,__LINE__,__FILE__);
    abort();
  }
  float fproc = (float)gid * (float)npe_ / (float)nglobal + 0.5;
  proc = (int)fproc; 
  if( proc > npe_ ) proc = npe_;
  while( gid < proc_gnode_[proc] || gid >= proc_gnode_[proc+1] ) {
    if(  gid < proc_gnode_[proc] ) proc--;
    else proc++;
  }
  if(proc<0 || proc >= npe_){
    PetscPrintf( PETSC_COMM_SELF,
		 "%s line %d in %s ERROR: proc<0 || proc >= np\n",
		 __FUNCT__,__LINE__,__FILE__);
    abort();
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::resetMatrixAndVector"
/* Prometheus_LinSysCore::resetMatrixAndVector() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::resetMatrixAndVector( double s ) 
{
  PRINTD(s); 
  int ierr;
  
  // zero data 
  ierr = resetRHSVector(s); CHKERRQ(ierr); 
  ierr = resetMatrix(s); CHKERRQ(ierr); 

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::resetMatrix"
/* Prometheus_LinSysCore::resetMatrix() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::resetMatrix( double s ) 
{
  PRINTD(s);
  if( s != 0.0 ){ SETERRQ(1, "Prometheus: s != 0." ); }
  PromMatrix *KK = prom_->grid0_->stiffness_;
  if( KK == NULL ) { SETERRQ(1, "Prometheus: no stifness matrix" ); }
  int ierr = KK->ZeroEntries(); CHKERRQ(ierr); 

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::resetRHSVector"
/* Prometheus_LinSysCore::resetRHSVector() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::resetRHSVector( double s ) 
{
  PRINT1(currentRHS_); 

  //
  if( bVecArr_ != NULL ) {
    PromVector *BB = bVecArr_[currentRHS_];         assert(currentRHS_>=0);
    if( BB == NULL ) { SETERRQ(1,"Prometheus: BB == NULL"); }
    int ierr = BB->Set( s ); CHKERRQ(ierr);
  }
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setNumRHSVectors"
/* Prometheus_LinSysCore::setNumRHSVectors() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::setNumRHSVectors( int numrhss, const int* rhsids ) 
{
  PRINT1(numrhss); 
  int ierr, ii, oldnumrhs = numRHSs_;

  PetscFunctionBegin;
  numRHSs_ = numrhss;
  if ( numrhss == oldnumrhs ) /* void */ ; // noop
  else if( numrhss == 0 ){ // deleting ???
    if( rhsIDs_ == NULL ){ SETERRQ(1,"Prometheus: rhsIDs == NULL");} 
    PetscFree( rhsIDs_ ); rhsIDs_ = NULL; 
    for( ii = 0 ; ii < oldnumrhs ; ii++ ){
      PromVector *BB = bVecArr_[ii]; 
      assert(BB!=NULL); delete BB; bVecArr_[ii] = NULL; 
    }
    PetscFree( bVecArr_ ); bVecArr_ = NULL;
  }
  else if( oldnumrhs > numrhss ) { // reducing 
    for( ii = numrhss ; ii < oldnumrhs ; ii++ ){
      PromVector *BB = bVecArr_[ii]; bVecArr_[ii] = NULL; delete BB;
      if( currentRHS_ == ii ) currentRHS_ = 0;
    }
  }
  else { // increasing
    if( rhsIDs_ != NULL ) PetscFree( rhsIDs_ );
    ierr = PetscMalloc( sizeof(int)*numRHSs_, &rhsIDs_ ); CHKERRQ(ierr);
    PromVector** oldBarr = bVecArr_;
    ierr = PetscMalloc(numRHSs_*sizeof(PromVector*),&bVecArr_);
    CHKERRQ(ierr);
    for(ii = 0 ; ii < oldnumrhs ; ii++) bVecArr_[ii] = oldBarr[ii]; 
    if( oldBarr != NULL ) PetscFree( oldBarr ); 
    ii = oldnumrhs;
    for( ii = oldnumrhs ; ii < numRHSs_ ; ii++ ) { // make new
      PromGrid *grid0 = prom_->grid0_; 
      if( grid0 != NULL ){
	PromVector *bb = NULL;
	if( numGlobalRowsLm_ == 0 ) bb = new PromCRVector( *grid0 );
	else if( use_complex_ ) bb = new PromComplexVector( *grid0 ); 
	else{ assert(lmap_!=NULL); bb = new PromCRVector(*grid0, lmap_);}
	ierr = bb->Set( 0.0 );  CHKERRQ(ierr);
	bVecArr_[ii] = bb;
      } 
      else bVecArr_[ii] = NULL; // called without initial setup (FEI?)
    }
  }
  for( ii = 0; ii < numRHSs_; ii++ ) rhsIDs_[ii] = rhsids[ii];   
  PetscFunctionReturn(0);
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setRHSID"
/* Prometheus_LinSysCore::setRHSID() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::setRHSID( int rhsID ) 
{
  if(verbose_>2)PRINT1(rhsID); 

  if( currentRHS_ < 0 || rhsID != rhsIDs_[currentRHS_] ) {
    for(int i=0; i<numRHSs_; i++){
      if( rhsIDs_[i] == rhsID ){
	currentRHS_ = i; assert(currentRHS_>=0);
	return 0;
      }
    }
    PetscPrintf(PETSC_COMM_SELF,"[%d]%s line %d in %s ERROR: ID not found\n",
		mype_,__FUNCT__,__LINE__,__FILE__);
    return 1;
  }
  else return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::formResidual"
/* Prometheus_LinSysCore::formResidual() ***************************
 * 
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::formResidual( double* values, int len )
{
  return formResidual_private( values, len, FALSE );
}

int Prometheus_LinSysCore::getResidual_i( double* values, int len )
{
  return formResidual_private( values, len, TRUE );
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::formResidual_private"
/* Prometheus_LinSysCore::formResidual_private() *********************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::formResidual_private(double* values, int len,
						const bool image/*=f*/)
{
  // explicit residual
  PRINT1(len);
  int ierr,eq,jj;  assert(currentRHS_>=0);
  PromVector *BB = bVecArr_[currentRHS_]; 
  PromCRVector *W = dynamic_cast<PromCRVector*>(work_); 
  double *pv; 
  const int my0prim = proc_primEq_[mype_], next0prim=proc_primEq_[mype_+1];
  const int nlocPrims = next0prim - my0prim; // first local Lm eq
  const int nlocLm = proc_lmEq_[mype_+1] - proc_lmEq_[mype_],iof=(image?1:0);
  PromMatrix *KK = prom_->grid0_->stiffness_;
  PromVector *XX = xVec_; assert(XX!=NULL);
  const int fact = (XX->isComplex() ? 2 : 1);
  
  if(BB == NULL){SETERRQ(1,"Prometheus: no RHS vector");}
  if(len != nlocPrims + nlocLm)SETERRQ(1,"Prometheus: len != numLocalRows_");
  // r <- K u -- manual residual hack
  if( !image ) {
    ierr = LSC_Residual( KK, BB, XX, work_ );    CHKERRQ(ierr);
  }

  // prim
  ierr = work_->GetArray( &pv );    CHKERRQ(ierr);
  for( eq = jj = 0 ; jj < nlocPrims ; jj++, eq += fact ){
    values[jj] = pv[eq+iof]; // b - Ax
  }
  ierr = work_->RestoreArray( &pv ); CHKERRQ(ierr);
  // lm
  if( W != NULL ) {
    ierr = W->p_->GetArray( &pv );    CHKERRQ(ierr);
    for( jj = 0 ; jj < nlocLm ; jj++, eq++ ){
      values[eq] = pv[jj]; // b - Ax 
    }
    ierr = W->p_->RestoreArray( &pv ); CHKERRQ(ierr);
  }
  else if( eq < len ) {
    while( eq < len ) values[eq++] = 0.0; // projection w/o any uzawa
  }
  if( eq != len ) SETERRQ2(6,"Prometheus: eq != len %d %d",eq,len);

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::LSC_Residual"
/* Prometheus_LinSysCore::LSC_Residual() ***************************
 *
 *   RR <- BB - K * XX
 *
 *  INPUT: 
 *   - KK:
 *   - BB:
 *   - XX:
 *   - RR: (out)
 *
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::LSC_Residual( PromMatrix *KK, 
					 const PromVector *const BB, 
					 PromVector *const XX,
					 PromVector *const RR )
{
  int ierr;
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); 
  PromMatrix *CC = use_uzawa_ ? uzawaC_ : (K==NULL ? NULL : K->C_);
  PromCRVector *R = dynamic_cast<PromCRVector*>(RR);
  PromCRVector *X = dynamic_cast<PromCRVector*>(XX);

  // scale
  if( prom_->right_ != NULL ) { 
    // scale x <- x * D^{1/2}
    ierr = XX->PromVector::PointwiseDiv( XX, prom_->right_ ); 
    CHKERRQ(ierr);
  }

  // rr =: Ax - b 
  ierr = KK->PromMatrix::Mult( XX, RR ); CHKERRQ(ierr);

  if( CC != NULL ){
    ierr = CC->MultTransposeAdd( X->p_, RR, RR ); CHKERRQ(ierr);
    assert( R->p_ != NULL );
    ierr = CC->Mult( X, R->p_ );             CHKERRQ(ierr);
  }
  else if( R && R->p_ != NULL ) R->p_->Set( 0.0 ); 
  // proj
  const PromSolver *solver = prom_->solver_;  assert(solver != NULL);
  if( solver->projector_ != NULL ) { 
    ierr = solver->projector_->Apply_Ct_add( X->p_, R, R ); CHKERRQ(ierr);
  }
  // unscale
  if( prom_->right_ != NULL ) { 
    // scale r <- r * L^{1/2} ( L^{-1/2} A R^{-1/2} R^{1/2} x)
    ierr = XX->PromVector::PointwiseMult(XX, prom_->right_);
    CHKERRQ(ierr);
  }
  if( prom_->left_ != NULL ) { 
    // scale r <- r * L^{1/2} ( L^{-1/2} A R^{-1/2} R^{1/2} x)
    ierr = RR->PromVector::PointwiseDiv( RR, prom_->left_ ); 
    CHKERRQ(ierr);
  }
  // w = b - Ax.
  ierr = RR->AYPX( -1.0, BB ); CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::getSolution"
/* Prometheus_LinSysCore::getSolution() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getSolution( double* answers, int len )
{
  return getSolution_private( answers, len, FALSE );
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::getSolution_i"
/* Prometheus_LinSysCore::getSolution_i() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getSolution_i( double* answers, int len )
{
  return getSolution_private( answers, len, TRUE );
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::getSolution_private"
/* Prometheus_LinSysCore::getSolution_private() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getSolution_private( double* answers, int len, 
						const bool image )
{
  //
  //The caller must allocate the memory for 'answers' and eqnNumbers,
  //and len must be set to the right value -- i.e., len should equal
  //numLocalRows_.
  // 
  PRINT;
  if (len != proc_globalEq_[mype_+1] - proc_globalEq_[mype_]) {
    SETERRQ(1,"Prometheus: len != numLocalRows_");
  }
  int ierr, jj, ii;
  const int my0prim = proc_primEq_[mype_], next0prim=proc_primEq_[mype_+1];
  const int nlocPrims = next0prim - my0prim;
  const int nlocLm = proc_lmEq_[mype_+1] - proc_lmEq_[mype_];
  double *vv; PromVector *XX = xVec_; assert(XX!=NULL);
  if(len != proc_globalEq_[mype_+1] - proc_globalEq_[mype_]){
    SETERRQ(1,"Prometheus: len != numLocalRows_");
  }
  const int fact = XX->isComplex() ? 2 : 1, ioff = image ? 1 : 0;
  ierr = XX->GetArray( &vv ); CHKERRQ(ierr);
  for( ii = 0, jj = ioff ; ii < nlocPrims ; ii++, jj += fact ) {
    answers[ii] = vv[jj];
  }
  ierr = XX->RestoreArray( &vv );  CHKERRQ(ierr);

  // lm
  PromCRVector *X = dynamic_cast<PromCRVector*>(XX); 
  if( X != NULL ) {
    PromVector *pp = X->p_; assert(pp != NULL);    
    ierr = pp->GetArray( &vv );    CHKERRQ(ierr);
    for( jj = 0 ; jj < nlocLm ; jj++, ii++ ){
      answers[ii] = vv[jj]; // b - Ax = -(Ax - b)
    }
    ierr = pp->RestoreArray( &vv ); CHKERRQ(ierr);
  }
  if( ii != proc_globalEq_[mype_+1] - proc_globalEq_[mype_]){
    SETERRQ(1,"Prometheus: ii != numLocalRows_");
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::getSolnEntry"
/* Prometheus_LinSysCore::getSolnEntry() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getSolnEntry( int eqnNumber, double& answer ) 
{
  //
  //This function returns a single solution entry, the coefficient for
  //equation number eqnNumber.
  //
  if( verbose_>3 ) PRINT;
  if(eqnNumber<proc_globalEq_[mype_] || eqnNumber>=proc_globalEq_[mype_+1]){
    SETERRQ(1,"eqnNumber out of range");
  }
  int ierr, myeq, proc; bool lamb;
  const int my0prim = proc_primEq_[mype_], my0Lm = proc_lmEq_[mype_];
  double *vv;
  ierr = eq_map_private( eqnNumber, &myeq, &lamb, &proc ); CHKERRQ(ierr);
  if( proc != mype_ ){ SETERRQ(1,"pe != mype");}
  PromVector *XX = xVec_;  assert(XX!=NULL);
  if( !lamb ) {
    ierr = XX->GetArray( &vv ); CHKERRQ(ierr);
    int id = (myeq - my0prim)*(XX->isComplex() ? 2 : 1);
    answer = vv[id]; // causes an error in salinas!!!
    ierr = XX->RestoreArray( &vv ); CHKERRQ(ierr);
  }
  else {
    PromCRVector *X = dynamic_cast<PromCRVector*>(XX); assert(X!=NULL);
    if( X->p_ == NULL ) {
      answer = 0.0; // all projections
      SETERRQ(1,"Lagrange multiplier requested for projection eq");
    }
    else {
      ierr = X->p_->GetArray( &vv ); CHKERRQ(ierr);
      answer = vv[myeq - my0Lm]; 
      ierr = X->p_->RestoreArray( &vv ); CHKERRQ(ierr);
    }
  }

  return 0;
}
 
#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::getSolnEntry_i"
/* Prometheus_LinSysCore::getSolnEntry_i() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getSolnEntry_i( int eqnNumber, double& answer )
{
  //
  //This function returns a single solution entry, the coefficient for
  //equation number eqnNumber.
  //
  if(verbose_>3) PRINT;
  if(eqnNumber<proc_globalEq_[mype_] || eqnNumber>=proc_globalEq_[mype_+1]){
    SETERRQ(1,"eqnNumber out of range");
  }
  int ierr, myeq, proc; bool lamb;
  const int my0prim = proc_primEq_[mype_], my0Lm = proc_lmEq_[mype_];
  double *vv;
  ierr = eq_map_private( eqnNumber, &myeq, &lamb, &proc ); CHKERRQ(ierr);
  if( proc != mype_ ){ SETERRQ(1,"pe != mype");}
  PromVector *XX = xVec_;  assert(XX!=NULL);
  if( !lamb ) {
    ierr = XX->GetArray( &vv ); CHKERRQ(ierr);
    answer = vv[2*(myeq-my0prim)+1]; 
    ierr = XX->RestoreArray( &vv ); CHKERRQ(ierr);
  }
  else {
    PromCRVector *X = dynamic_cast<PromCRVector*>(XX); assert(X!=NULL);
    if( X->p_ == NULL ) answer = 0.0; // all projections
    else {
      ierr = X->p_->GetArray( &vv ); CHKERRQ(ierr);
      answer = vv[myeq - my0Lm ]; 
      ierr = X->p_->RestoreArray( &vv ); CHKERRQ(ierr);
    }
  }

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::getMatrixRowLength"
/* Prometheus_LinSysCore::getMatrixRowLength() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getMatrixRowLength( int row, int& length )
{
  int ierr, myeq, proc, geq = row, nc, *ids; bool lamb;
  PromMatrix   *KK = prom_->grid0_->stiffness_;
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); 
  PromMatrix   *CC = use_uzawa_ ? uzawaC_ :(K==NULL ? NULL : K->C_);
  
  ierr = eq_map_private( geq, &myeq, &lamb, &proc ); CHKERRQ(ierr);
  if( !lamb ) {
    if(proc != mype_)SETERRQ1(1,"proc!=myproc,proc=%d",proc);
    ierr = KK->GetRow( myeq, &nc, &ids );   CHKERRQ(ierr);
    ierr = KK->RestoreRow( myeq, &nc, &ids ); CHKERRQ(ierr);
    length = nc;
  }
  else {
    // projC_ -- todo!
    if( CC == NULL ) SETERRQ1(1,"C==NULL, row=%d", row );
    if(proc != mype_) SETERRQ1(1,"proc!=myproc,proc=%d",proc);
    ierr = CC->GetRow( myeq, &nc, &ids );   CHKERRQ(ierr);
    ierr = CC->RestoreRow( myeq, &nc, &ids ); CHKERRQ(ierr);
    length = nc;
  }
  
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::getMatrixRow"
/* Prometheus_LinSysCore::getMatrixRow() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS: -- note A could have regularization
 */
int Prometheus_LinSysCore::getMatrixRow( int row, double* coefs, int* indices,
					 int len, int& rowLength)
{
  int ierr, myeq, proc, geq = row, nc, *ids; double *vv; bool lamb;
  PromMatrix *KK = prom_->grid0_->stiffness_;
  PromCRMatrix *K = dynamic_cast<PromCRMatrix*>(KK); 
  PromMatrix *CC = use_uzawa_ ? uzawaC_ : (K==NULL ? NULL : K->C_);
  
  ierr = eq_map_private( geq, &myeq, &lamb, &proc ); CHKERRQ(ierr);
  if( !lamb ) {
    if(proc != mype_) SETERRQ1(1,"proc!=myproc, proc=%d",proc);
    ierr = KK->GetRow( myeq, &nc, &ids, &vv );   CHKERRQ(ierr);
    if(nc>len)SETERRQ2(1,"nc>len, nc=%d len=%d",nc,len);
    for(int ii=0;ii<nc;ii++){ coefs[ii] = vv[ii]; indices[ii] = ids[ii]; }
    ierr = KK->RestoreRow( myeq, &nc, &ids, &vv ); CHKERRQ(ierr);
    rowLength = nc;
  }
  else {
    // projC_ -- todo!
    if( CC == NULL ) SETERRQ1(1,"C==NULL, row=%d", row );
    if(proc!=mype_)SETERRQ1(1,"proc!=myproc, proc=%d",proc);
    ierr = CC->GetRow( myeq, &nc, &ids, &vv );   CHKERRQ(ierr);
    if(nc>len)SETERRQ2(1,"nc>len, nc=%d len=%d",nc,len);
    for(int ii=0;ii<nc;ii++){ coefs[ii] = vv[ii]; indices[ii] = ids[ii]; }
    ierr = CC->RestoreRow( myeq, &nc, &ids, &vv ); CHKERRQ(ierr);
    rowLength = nc;
  }
  
  return 0;
}

/***********************************************************************
 *                      Prometheus_EssBCData
 **********************************************************************/
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_EssBCData::Prometheus_EssBCData"
/* Prometheus_EssBCData::Prometheus_EssBCData() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
Prometheus_EssBCData::Prometheus_EssBCData( int* globalEqn,
					    double* alph,
					    double* gamm,
					    int len)
{
  len_ = len; int ii;
  PetscMalloc(len*sizeof(int),&globalEqn_); 
  if( !globalEqn_ ){ NOPROM_PRINTS("memory error"); abort(); }
  PetscMalloc(len*sizeof(double),&alpha_); 
  if( !alpha_ ){ NOPROM_PRINTS("memory error"); abort(); }
  PetscMalloc(len*sizeof(double),&gamma_); 
  if( !gamma_ ){ NOPROM_PRINTS("memory error"); abort(); }
  for(ii=0;ii<len;ii++){
    globalEqn_[ii] = globalEqn[ii]; alpha_[ii] = alph[ii]; 
    gamma_[ii] = gamm[ii];
  }
  next_ = NULL;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_EssBCData::~Prometheus_EssBCData"
/* Prometheus_EssBCData::~Prometheus_EssBCData() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
Prometheus_EssBCData::~Prometheus_EssBCData()
{
  PetscFree(globalEqn_); PetscFree(alpha_); PetscFree(gamma_);
}

/***********************************************************************
 *                      Prometheus_OthBCData
 **********************************************************************/
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_OthBCData::Prometheus_OthBCData"
/* Prometheus_OthBCData::Prometheus_OthBCData() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
Prometheus_OthBCData::Prometheus_OthBCData( int* globalEqn, 
					    double* alph, 
					    double* bet, 
					    double* gamm,
					    int len) 
{ 
  len_ = len; int ii;
  PetscMalloc(len*sizeof(int),&globalEqn_);
  if( !globalEqn_ ){ NOPROM_PRINTS("memory error"); return; }
  PetscMalloc(len*sizeof(double),&alpha_);
  if( !alpha_ ){ NOPROM_PRINTS("memory error"); return; }
  PetscMalloc(len*sizeof(double),&beta_);
  if( !beta_ ){ NOPROM_PRINTS("memory error"); return; }
  PetscMalloc(len*sizeof(double),&gamma_);
  if( !gamma_ ){ NOPROM_PRINTS("memory error"); return; }
  for(ii=0;ii<len;ii++){
    globalEqn_[ii] = globalEqn[ii]; alpha_[ii] = alph[ii]; 
    beta_[ii] = bet[ii]; gamma_[ii] = gamm[ii]; 
  }
  next_ = NULL;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_OthBCData::Prometheus_OthBCData"
/* Prometheus_OthBCData::Prometheus_OthBCData() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
Prometheus_OthBCData::~Prometheus_OthBCData()
{
  PetscFree(globalEqn_); PetscFree(alpha_); 
  PetscFree(gamma_); PetscFree(beta_);
}

/* matrix manipulation functions ***************************************
 *
 */
#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::getMatrixPtr"
/* Prometheus_LinSysCore::getMatrixPtr() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getMatrixPtr( Data &data )
{
  //PRINT;
  data.setTypeName(matrix_name);
  PromMatrix *KK = prom_->grid0_->stiffness_;
  data.setDataPtr( (void*)KK );

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::copyInMatrix"
/* Prometheus_LinSysCore::copyInMatrix() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::copyInMatrix( double scalar, const Data &data ) 
{
  //
  //Overwrites the current internal matrix with a scaled copy of the
  //input argument.
  //
  PRINTD(scalar);
  int ierr;
  if( strcmp(matrix_name, data.getTypeName()) != 0 ){
    SETERRQ(1,"data's type string not 'Prometheus_Matrix'.");
  }
  const PromMatrix *const inA = (PromMatrix*)data.getDataPtr(); 
  PromMatrix *KK = prom_->grid0_->stiffness_;
  
  // better have same non-zero pattern!!! -- DIFFERENT_NONZERO_PATTERN
  ierr = KK->ZeroEntries(); CHKERRQ(ierr);
  ierr = KK->AXPY( scalar, inA );  CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::copyOutMatrix"
/* Prometheus_LinSysCore::copyOutMatrix() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::copyOutMatrix( double scalar, Data& data ) 
{
  //
  //Passes out a scaled (created) copy of the current internal matrix.
  //
  PRINTD(scalar); int ierr; 
  PromMatrix *outmat, *KK = prom_->grid0_->stiffness_;

  // collective -- this should not be needed (maybe)
  ierr = KK->Assembly( );   CHKERRQ(ierr);
  outmat = new PromMatrix( KK );
  if(outmat==NULL || !outmat->isOK())SETERRQ(1,"new PromMatrix failed");
  //ierr = newA->Copy( A );  CHKERRQ(ierr);  // construct instead!!!
  if( scalar != 1.0 ) {
    ierr = outmat->Scale( scalar );  CHKERRQ(ierr);
  }

  data.setTypeName(matrix_name);  
  data.setDataPtr( (void*)outmat );

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::sumInMatrix"
/* Prometheus_LinSysCore::sumInMatrix() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumInMatrix( double scalar, const Data& data ) 
{
  PRINTD(scalar);
  int ierr;  

  if( strcmp(matrix_name, data.getTypeName()) != 0 ){
    SETERRQ(1,"data's type string not 'Prometheus_Matrix'.");
  }
  PromMatrix *const inK = (PromMatrix*)data.getDataPtr(); 
  PromMatrix *KK = prom_->grid0_->stiffness_;
  
  // collective
  ierr = inK->Assembly();   CHKERRQ(ierr);
  //   Computes Y = a*X + Y. int MatAXPY(Scalar *a,Mat X,Mat Y)
  prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FEI_MAT_AXPY]);
  ierr = KK->AXPY( scalar, inK );  CHKERRQ(ierr);
  prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FEI_MAT_AXPY]);

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::getRHSVectorPtr"
/* Prometheus_LinSysCore::getRHSVectorPtr() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::getRHSVectorPtr( Data& data ) 
{
  PRINT;
  
  data.setTypeName( vector_name ); assert(currentRHS_>=0);
  data.setDataPtr( (void*)bVecArr_[currentRHS_] );

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::copyInRHSVector"
/* Prometheus_LinSysCore::copyInRHSVector() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::copyInRHSVector( double scalar, const Data& data ) 
{
  //
  //Overwrites the current internal matrix with a scaled copy of the
  //input argument.
  //
  PRINTD(scalar);

  int ierr;
  if( strcmp(vector_name, data.getTypeName()) != 0 ){
    SETERRQ(1,"data's type string not 'Prometheus_Vector'.");
  }

  assert(currentRHS_>=0);
  const PromVector *const yy = (PromVector*)data.getDataPtr();
  const PromCRVector *const Y = dynamic_cast<const PromCRVector*>(yy);
  PromVector *const BB = bVecArr_[currentRHS_];
  PromCRVector *const B = dynamic_cast<PromCRVector*>(BB);

  if( scalar != 1.0 ) {
    ierr =  BB->Set( 0.0 ); CHKERRQ(ierr);
    ierr =  BB->AXPY( scalar, yy );  CHKERRQ(ierr);
  }
  else {
    ierr = BB->PromVector::Copy( yy );  CHKERRQ(ierr);
  }

  if( B && B->p_ != NULL ){
    assert(Y->p_ != NULL);
    if( scalar != 1.0 ) {
      ierr =  B->p_->Set( 0.0 ); CHKERRQ(ierr);
      ierr =  B->p_->AXPY( scalar, Y->p_ );  CHKERRQ(ierr);
    }
    else {
      ierr = B->p_->Copy( Y->p_ );  CHKERRQ(ierr);
    }
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::copyOutRHSVector"
/* Prometheus_LinSysCore::copyOutRHSVector() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::copyOutRHSVector( double scalar, Data& data ) 
{ 
  PRINTD(scalar);
  int ierr; assert(currentRHS_>=0);
  PromVector *BB = bVecArr_[currentRHS_],*outvec;
  if( BB->isComplex() ) {
    outvec = new PromComplexVector( BB ); CHKNEW(outvec,1); 
  }
  else {
    outvec = new PromVector( BB ); CHKNEW(outvec,1); 
  }
  ierr = outvec->Copy( BB );  CHKERRQ(ierr);
  if( scalar != 1.0 ) {
    ierr = outvec->Scale( scalar );  CHKERRQ(ierr);
  }
  //if( B->p_ != NULL ) assert(outvec->p_ != NULL);  
  
  data.setTypeName( vector_name );
  data.setDataPtr( (void*)outvec );  

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::sumInRHSVector"
/* Prometheus_LinSysCore::sumInRHSVector() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::sumInRHSVector( double scalar, const Data& data ) 
{
  PRINTD(scalar);
  int ierr;
  if( strcmp(vector_name, data.getTypeName()) != 0 ){
    SETERRQ(1,"data's type string not 'Prometheus_Vector'.");
  }
  assert(currentRHS_>=0);
  const PromVector *const XX = (PromVector*)data.getDataPtr();
  const PromCRVector *X = dynamic_cast<const PromCRVector*>(XX); 
  PromVector *const BB = bVecArr_[currentRHS_];
  PromCRVector *const B = dynamic_cast<PromCRVector*>(BB);
  
  // Computes y = alpha x + y. 
  ierr = BB->AXPY( scalar, XX );  CHKERRQ(ierr);
  if( B && B->p_ != NULL ){ 
    ierr = B->p_->AXPY( scalar, X->p_ );  CHKERRQ(ierr); 
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::destroyMatrixData"
/*Prometheus_LinSysCore::destroyMatrixData() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::destroyMatrixData( Data& data ) 
{
  PRINT;

  if( strcmp(matrix_name, data.getTypeName()) != 0 ){
    SETERRQ(1,"data's type string not 'Prometheus_Matrix'.");
  }
  PromMatrix *inA = (PromMatrix*)data.getDataPtr(); 
  PromMatrix *KK = prom_->grid0_->stiffness_;
  if( inA == KK ) {
    SETERRQ(1,"Dstroying system matrix");
  }

  delete inA; 
  data.setDataPtr( NULL );   

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::destroyVectorData"
/* Prometheus_LinSysCore::destroyVectorData() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::destroyVectorData( Data& data ) 
{
  PRINT;

  if( strcmp(vector_name, data.getTypeName()) != 0 ){
    SETERRQ(1,"data's type string not 'Prometheus_Vector'.");
  }
  PromVector *XX = (PromVector*)data.getDataPtr();

  delete XX; 
  data.setDataPtr( NULL );  

  return 0;
}

/*** NOT implemetnted ********************************************************/

#undef __FUNCT__    
#define __FUNCT__ "Prometheus_LinSysCore::setStiffnessMatrices"
/* Prometheus_LinSysCore::setStiffnessMatrices() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::setStiffnessMatrices(GlobalID elemBlock,
						int numElems,
						const GlobalID* elemIDs,
						const double *const*const *st,
						int numEqnsPerElem,
						const int *const * eqnIndices)
{
  return 0;
}

#undef __FUNCT__   
#define __FUNCT__ "Prometheus_LinSysCore::setLoadVectors"
/* Prometheus_LinSysCore::setLoadVectors() ***************************
 *
 *  - 
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::setLoadVectors(GlobalID elemBlock,
					  int numElems,
					  const GlobalID* elemIDs,
					  const double *const *load,
					  int numEqnsPerElem,
					  const int *const * eqnIndices)
{
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "Prometheus_LinSysCore::writeSystem"
/* Prometheus_LinSysCore::writeSystem() ***************************
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 */
int Prometheus_LinSysCore::writeSystem( const char* name )
{
  // easy to do with PETSc
  SETERRQ(1,"writeSystem not implemented");
}
